# -*- coding: utf-8 -*-
import logging
from enum import Enum
from typing import List, Optional

import numpy as np
import pandas as pd
import plotly.graph_objs as go
from plotly.subplots import make_subplots

from zvtm.contract.api import decode_entity_id
from zvtm.contract.data_type import Bean
from zvtm.contract.normal_data import NormalData
from zvtm.utils import to_string
from zvtm.utils.pd_utils import pd_is_not_null

logger = logging.getLogger(__name__)


class ChartType(Enum):
    kline = 'kline'
    line = 'line'
    area = 'area'
    scatter = 'scatter'
    histogram = 'histogram'
    pie = 'pie'


@to_string
class Rect(Bean):

    def __init__(self, x0=None, y0=None, x1=None, y1=None) -> None:
        # left-bottom
        self.x0 = x0
        self.y0 = y0
        # right-top
        self.x1 = x1
        self.y1 = y1


class Draw(object):
    def draw_kline(self, width=None, height=None, title=None, keep_ui_state=True, show=False, **kwargs):
        return self.draw(ChartType.kline, width=width, height=height, title=title, keep_ui_state=keep_ui_state,
                         show=show,
                         **kwargs)

    def draw_line(self, width=None, height=None, title=None, keep_ui_state=True, show=False, **kwargs):
        return self.draw_scatter(mode='lines', width=width, height=height, title=title,
                                 keep_ui_state=keep_ui_state, show=show, **kwargs)

    def draw_area(self, width=None, height=None, title=None, keep_ui_state=True, show=False, **kwargs):
        return self.draw_scatter(mode='none', width=width, height=height, title=title,
                                 keep_ui_state=keep_ui_state, show=show, **kwargs)

    def draw_scatter(self, mode='markers', width=None, height=None,
                     title=None, keep_ui_state=True, show=False, **kwargs):
        return self.draw(ChartType.scatter, mode=mode, width=width, height=height, title=title,
                         keep_ui_state=keep_ui_state,
                         show=show, **kwargs)

    def draw_histogram(self, width=None, height=None, title=None, keep_ui_state=True, show=False, **kwargs):
        return self.draw(ChartType.histogram, width=width, height=height, title=title, keep_ui_state=keep_ui_state,
                         show=show, **kwargs)

    def draw_pie(self, width=None, height=None, title=None, keep_ui_state=True, show=False, **kwargs):
        return self.draw(ChartType.pie, width=width, height=height, title=title, keep_ui_state=keep_ui_state,
                         show=show, **kwargs)

    def draw(self,
             main_chart=ChartType.kline,
             sub_chart='bar',
             mode='lines',
             width=None,
             height=None,
             title=None,
             keep_ui_state=True,
             show=False,
             **kwargs):

        raise NotImplementedError()

    def default_layout(self,
                       main_chart=None,
                       width=None,
                       height=None,
                       title=None,
                       keep_ui_state=True,
                       **layout_params):
        if keep_ui_state:
            uirevision = True
        else:
            uirevision = None

        if main_chart == ChartType.histogram:
            xaxis = None
        else:
            xaxis = dict(
                linecolor="#BCCCDC",
                showgrid=False,
                showspikes=True,  # Show spike line for X-axis
                # Format spike
                spikethickness=2,
                spikedash="dot",
                spikecolor="#999999",
                spikemode="across",
                rangeselector=dict(
                    buttons=list([
                        dict(count=1,
                             label="1m",
                             step="month",
                             stepmode="backward"),
                        dict(count=3,
                             label="3m",
                             step="month",
                             stepmode="backward"),
                        dict(count=6,
                             label="6m",
                             step="month",
                             stepmode="backward"),
                        dict(count=1,
                             label="YTD",
                             step="year",
                             stepmode="todate"),
                        dict(count=1,
                             label="1y",
                             step="year",
                             stepmode="backward"),
                        dict(step="all")
                    ])
                ),
                rangeslider=dict(
                    visible=True,
                ),
                type="date"
            )

        return dict(showlegend=True,
                    plot_bgcolor="#FFF",
                    hovermode="x",
                    hoverdistance=100,  # Distance to show hover label of data point
                    spikedistance=1000,  # Distance to show spike
                    uirevision=uirevision,
                    height=height,
                    width=width,
                    title=title,
                    yaxis=dict(
                        autorange=True,
                        fixedrange=False,
                        zeroline=False,
                        linecolor="#BCCCDC",
                        showgrid=False,
                    ),
                    xaxis=xaxis,
                    legend_orientation="h",
                    hoverlabel={"namelength": -1},
                    **layout_params)


class Drawable(object):

    def drawer(self):
        drawer = Drawer(main_df=self.drawer_main_df(),
                        main_data=self.drawer_main_data(),
                        factor_df_list=self.drawer_factor_df_list(),
                        factor_data_list=self.drawer_factor_data_list(),
                        sub_df_list=self.drawer_sub_df_list(),
                        sub_data_list=self.drawer_sub_data_list(),
                        sub_col_chart=self.drawer_sub_col_chart(),
                        annotation_df=self.drawer_annotation_df(),
                        rects=self.drawer_rects())
        return drawer

    def draw(self,
             main_chart=ChartType.kline,
             width=None,
             height=None,
             title=None,
             keep_ui_state=True,
             show=False,
             **kwargs):
        return self.drawer().draw(main_chart=main_chart, width=width, height=height, title=title,
                                  keep_ui_state=keep_ui_state,
                                  show=show,
                                  **kwargs)

    def drawer_main_df(self) -> Optional[pd.DataFrame]:
        return None

    def drawer_main_data(self) -> Optional[NormalData]:
        return None

    def drawer_factor_df_list(self) -> Optional[List[pd.DataFrame]]:
        return None

    def drawer_factor_data_list(self) -> Optional[List[NormalData]]:
        return None

    def drawer_sub_df_list(self) -> Optional[List[pd.DataFrame]]:
        return None

    def drawer_sub_data_list(self) -> Optional[List[NormalData]]:
        return None

    def drawer_annotation_df(self) -> Optional[pd.DataFrame]:
        return None

    def drawer_rects(self) -> Optional[List[Rect]]:
        return None

    def drawer_sub_col_chart(self) -> Optional[dict]:
        return None


class StackedDrawer(Draw):
    def __init__(self, *drawers) -> None:
        super().__init__()
        assert len(drawers) > 1
        self.drawers: List[Drawer] = drawers

    def make_y_layout(self, index, total, start_index=1, domain_range=(0, 1)):
        part = (domain_range[1] - domain_range[0]) / total

        if index == 1:
            yaxis = 'yaxis'
            y = 'y'
        else:
            yaxis = f'yaxis{index}'
            y = f'y{index}'

        return yaxis, y, dict(anchor="x",
                              autorange=True,
                              fixedrange=False,
                              zeroline=False,
                              linecolor="#BCCCDC",
                              showgrid=False,
                              domain=[domain_range[0] + part * (index - start_index),
                                      domain_range[0] + part * (index - start_index + 1)])

    def draw(self,
             main_chart=ChartType.kline,
             sub_chart='bar',
             mode='lines',
             width=None,
             height=None,
             title=None,
             keep_ui_state=True,
             show=False,
             **kwargs):
        stacked_fig = go.Figure()

        total = len(self.drawers)
        start = 1
        domain_range = (0, 1)
        for drawer in self.drawers:
            if drawer.has_sub_plot():
                domain_range = (0.2, 1)
                start = 2
                break
        for index, drawer in enumerate(self.drawers, start=start):
            traces, sub_traces = drawer.make_traces(main_chart=main_chart, sub_chart=sub_chart, mode=mode, **kwargs)

            # fix sub traces as the bottom
            if sub_traces:
                yaxis, y, layout = self.make_y_layout(index=1, total=1, domain_range=(0, 0.2))
                # update sub_traces with yaxis
                for trace in sub_traces:
                    trace.yaxis = y
                stacked_fig.add_traces(sub_traces)
                stacked_fig.layout[yaxis] = layout

            # make y layouts
            yaxis, y, layout = self.make_y_layout(index=index, total=total, start_index=start,
                                                  domain_range=domain_range)

            stacked_fig.layout[yaxis] = layout

            # update traces with yaxis
            for trace in traces:
                trace.yaxis = y
            stacked_fig.add_traces(traces)

            # update shapes with yaxis
            if drawer.rects:
                for rect in drawer.rects:
                    stacked_fig.add_shape(type="rect",
                                          x0=rect.x0, y0=rect.y0, x1=rect.x1, y1=rect.y1,
                                          line=dict(
                                              color="RoyalBlue",
                                              width=1),
                                          # fillcolor="LightSkyBlue",
                                          yref=y)

            # annotations
            if pd_is_not_null(drawer.annotation_df):
                stacked_fig.layout['annotations'] = annotations(drawer.annotation_df, yref=y)

        stacked_fig.update_layout(
            self.default_layout(main_chart=main_chart, width=width, height=height, title=title,
                                keep_ui_state=keep_ui_state))

        if show:
            stacked_fig.show()
        else:
            return stacked_fig


class Drawer(Draw):
    def __init__(self,
                 main_df: pd.DataFrame = None,
                 factor_df_list: List[pd.DataFrame] = None,
                 sub_df_list: pd.DataFrame = None,
                 main_data: NormalData = None,
                 factor_data_list: List[NormalData] = None,
                 sub_data_list: NormalData = None,
                 sub_col_chart: Optional[dict] = None,
                 rects: List[Rect] = None,
                 annotation_df: pd.DataFrame = None) -> None:
        """

        :param main_df: df for main chart
        :param factor_df_list: list of factor df on main chart
        :param sub_df_list: df for sub chart under main chart
        :param main_data: NormalData wrap main_df,use either
        :param factor_data_list: list of NormalData wrap factor_df,use either
        :param sub_data_list: NormalData wrap sub_df,use either
        :param annotation_df:
        """

        # 主图数据
        if main_data is None:
            main_data = NormalData(main_df)
        self.main_data: NormalData = main_data

        # 主图因子
        if not factor_data_list and factor_df_list:
            factor_data_list = []
            for df in factor_df_list:
                factor_data_list.append(NormalData(df))
        # 每一个df可能有多个column, 代表多个指标，对于连续型的，可以放在一个df里面
        # 对于离散型的，比如一些特定模式的连线，放在多个df里面较好，因为index不同
        self.factor_data_list: List[NormalData] = factor_data_list

        # 副图数据
        if not sub_data_list and sub_df_list:
            sub_data_list = []
            for df in sub_df_list:
                sub_data_list.append(NormalData(df))
        # 每一个df可能有多个column, 代表多个指标，对于连续型的，可以放在一个df里面
        # 对于离散型的，比如一些特定模式的连线，放在多个df里面较好，因为index不同
        self.sub_data_list: List[NormalData] = sub_data_list

        # 幅图col对应的图形，line or bar
        self.sub_col_chart = sub_col_chart

        # 主图的标记数据
        self.annotation_df = annotation_df

        # list of rect
        self.rects = rects

    def add_factor_df(self, df: pd.DataFrame):
        self.add_factor_data(NormalData(df))

    def add_factor_data(self, data: NormalData):
        if not self.factor_data_list:
            self.factor_data_list = []
        self.factor_data_list.append(data)

    def add_sub_df(self, df: pd.DataFrame):
        self.add_sub_data(NormalData(df))

    def add_sub_data(self, data: NormalData):
        if not self.sub_data_list:
            self.sub_data_list = []
        self.sub_data_list.append(data)

    def has_sub_plot(self):
        return self.sub_data_list is not None and not self.sub_data_list[0].empty()

    def make_traces(self,
                    main_chart=ChartType.kline,
                    sub_chart='bar',
                    mode='lines',
                    yaxis='y',
                    **kwargs):
        traces = []
        sub_traces = []

        for entity_id, df in self.main_data.entity_map_df.items():
            df = df.select_dtypes(np.number)
            code = entity_id
            try:
                _, _, code = decode_entity_id(entity_id)
            except Exception:
                pass

            # 构造主图
            if main_chart == ChartType.kline:
                trace_name = '{}_kdata'.format(code)
                trace = go.Candlestick(x=df.index, open=df['open'], close=df['close'], low=df['low'], high=df['high'],
                                       name=trace_name, yaxis=yaxis, **kwargs)
                traces.append(trace)
            elif main_chart == ChartType.scatter:
                for col in df.columns:
                    trace_name = '{}_{}'.format(code, col)
                    ydata = df[col].values.tolist()
                    traces.append(go.Scatter(x=df.index, y=ydata, mode=mode, name=trace_name, yaxis=yaxis, **kwargs))
            elif main_chart == ChartType.histogram:
                for col in df.columns:
                    trace_name = '{}_{}'.format(code, col)
                    x = df[col].tolist()
                    trace = go.Histogram(
                        x=x,
                        name=trace_name,
                        **kwargs
                    )
                    traces.append(trace)
                    annotation = [dict(
                        entity_id=entity_id,
                        timestamp=x[-1],
                        value=0,
                        flag=f'{trace_name}:{x[-1]}',
                    )]
                    annotation_df = pd.DataFrame.from_records(annotation, index=['entity_id', 'timestamp'])
                    if pd_is_not_null(self.annotation_df):
                        self.annotation_df = pd.concat([self.annotation_df, annotation_df])
                    else:
                        self.annotation_df = annotation_df
            elif main_chart == ChartType.pie:
                for _, row in df.iterrows():
                    traces.append(go.Pie(name=entity_id, labels=df.columns.tolist(), values=row.tolist(), **kwargs))
                    break
            else:
                assert False

            # 构造主图指标
            if self.factor_data_list:
                for factor_data in self.factor_data_list:
                    if not factor_data.empty():
                        factor_df = factor_data.entity_map_df.get(entity_id)
                        factor_df = factor_df.select_dtypes(np.number)
                        if pd_is_not_null(factor_df):
                            for col in factor_df.columns:
                                trace_name = '{}_{}'.format(code, col)
                                ydata = factor_df[col].values.tolist()

                                line = go.Scatter(x=factor_df.index, y=ydata, mode=mode, name=trace_name, yaxis=yaxis,
                                                  **kwargs)
                                traces.append(line)

            # 构造幅图
            if self.has_sub_plot():
                for sub_data in self.sub_data_list:
                    sub_df = sub_data.entity_map_df.get(entity_id)
                    if pd_is_not_null(sub_df):
                        sub_df = sub_df.select_dtypes(np.number)
                        for col in sub_df.columns:
                            trace_name = '{}_{}'.format(code, col)
                            ydata = sub_df[col].values.tolist()

                            def color(i):
                                if i > 0:
                                    return 'red'
                                else:
                                    return 'green'

                            colors = [color(i) for i in ydata]

                            the_sub_chart = None
                            if self.sub_col_chart is not None:
                                the_sub_chart = self.sub_col_chart.get(col)
                            if not the_sub_chart:
                                the_sub_chart = sub_chart

                            if the_sub_chart == ChartType.line:
                                sub_trace = go.Scatter(x=sub_df.index, y=ydata, name=trace_name, yaxis='y2',
                                                       marker=dict(color=colors))
                            else:
                                sub_trace = go.Bar(x=sub_df.index, y=ydata, name=trace_name, yaxis='y2',
                                                   marker=dict(color=colors))
                            sub_traces.append(sub_trace)

        return traces, sub_traces

    def add_rects(self, fig, yaxis='y'):
        if self.rects:
            for rect in self.rects:
                fig.add_shape(type="rect",
                              x0=rect.x0, y0=rect.y0, x1=rect.x1, y1=rect.y1,
                              line=dict(color="RoyalBlue",
                                        width=1),
                              # fillcolor="LightSkyBlue"
                              )
            fig.update_shapes(dict(xref='x', yref=yaxis))

    def draw(self,
             main_chart=ChartType.kline,
             sub_chart='bar',
             mode='lines',
             width=None,
             height=None,
             title=None,
             keep_ui_state=True,
             show=False,
             **kwargs):
        yaxis = 'y'
        traces, sub_traces = self.make_traces(main_chart=main_chart, sub_chart=sub_chart, mode=mode, yaxis=yaxis,
                                              **kwargs)

        if sub_traces:
            fig = make_subplots(rows=2, cols=1, row_heights=[0.8, 0.2], vertical_spacing=0.08, shared_xaxes=True)
            fig.add_traces(traces, rows=[1] * len(traces), cols=[1] * len(traces))
            fig.add_traces(sub_traces, rows=[2] * len(sub_traces), cols=[1] * len(sub_traces))
        else:
            fig = go.Figure()
            fig.add_traces(traces)

        # 绘制矩形
        self.add_rects(fig, yaxis=yaxis)

        fig.update_layout(self.default_layout(main_chart=main_chart, width=width, height=height, title=title,
                                              keep_ui_state=keep_ui_state))

        if sub_traces:
            fig.update_layout(xaxis_rangeslider_visible=False)
            fig.update_layout(xaxis2_rangeslider_visible=True, xaxis2_rangeslider_thickness=0.1)
        # 绘制标志
        if pd_is_not_null(self.annotation_df):
            fig.layout['annotations'] = annotations(self.annotation_df, yref=yaxis)

        if show:
            fig.show()
        else:
            return fig

    def draw_table(self, width=None, height=None, title=None, keep_ui_state=True, **kwargs):
        cols = self.main_data.data_df.index.names + self.main_data.data_df.columns.tolist()

        index1 = self.main_data.data_df.index.get_level_values(0).tolist()
        index2 = self.main_data.data_df.index.get_level_values(1).tolist()
        values = [index1] + [index2] + [self.main_data.data_df[col] for col in self.main_data.data_df.columns]

        data = go.Table(
            header=dict(values=cols,
                        fill_color=['#000080', '#000080'] + ['#0066cc'] * len(self.main_data.data_df.columns),
                        align='left',
                        font=dict(color='white', size=13)),
            cells=dict(values=values, fill=dict(color='#F5F8FF'), align='left'), **kwargs)

        fig = go.Figure()
        fig.add_traces([data])
        fig.update_layout(self.default_layout(width=width, height=height, title=title, keep_ui_state=keep_ui_state))

        fig.show()


def annotations(annotation_df: pd.DataFrame, yref='y'):
    """
    annotation_df format:
                                    value    flag    color
    entity_id    timestamp

    :param annotation_df:
    :param yref: specific yaxis e.g, y,y2,y3
    :return:

    """

    if pd_is_not_null(annotation_df):
        annotations = []
        for trace_name, df in annotation_df.groupby(level=0):
            if pd_is_not_null(df):
                for (_, timestamp), item in df.iterrows():
                    if 'color' in item:
                        color = item['color']
                    else:
                        color = '#ec0000'

                    value = round(item['value'], 2)
                    annotations.append(dict(
                        x=timestamp,
                        y=value,
                        xref='x',
                        yref=yref,
                        text=item['flag'],
                        showarrow=True,
                        align='center',
                        arrowhead=2,
                        arrowsize=1,
                        arrowwidth=2,
                        # arrowcolor='#030813',
                        ax=-10,
                        ay=-30,
                        bordercolor='#c7c7c7',
                        borderwidth=1,
                        bgcolor=color,
                        opacity=0.8
                    ))
        return annotations
    return None


if __name__ == '__main__':
    from zvtm.factors.zen import ZenFactor

    data_reader1 = ZenFactor(codes=['000338'], level='1d')
    data_reader2 = ZenFactor(codes=['000338'], level='1wk')
    print(data_reader2.data_df)

    stacked = StackedDrawer(data_reader1.drawer(), data_reader2.drawer()).draw_kline(show=True)
    # df = Stock1dHfqKdata.query_data(code='000338', start_timestamp='2015-01-01')
    # sub_df = FinanceFactor.query_data(code='000338', start_timestamp='2015-01-01',
    #                                   columns=[FinanceFactor.roe, FinanceFactor.entity_id, FinanceFactor.timestamp])
    #
    # Drawer(main_df=df, sub_df_list=[sub_df]).draw_kline(show=True)
# the __all__ is generated
__all__ = ['Rect', 'Draw', 'Drawable', 'StackedDrawer', 'Drawer', 'annotations']



# -*- coding: utf-8 -*-
from sqlalchemy import Column, String, DateTime, Float
from sqlalchemy.orm import declarative_base

from zvtm.contract import Mixin
from zvtm.contract.register import register_schema

DividendFinancingBase = declarative_base()


class DividendFinancing(DividendFinancingBase, Mixin):
    __tablename__ = 'dividend_financing'

    provider = Column(String(length=32))
    code = Column(String(length=32))

    # 分红总额
    dividend_money = Column(Float)

    # 新股
    ipo_issues = Column(Float)
    ipo_raising_fund = Column(Float)

    # 增发
    spo_issues = Column(Float)
    spo_raising_fund = Column(Float)
    # 配股
    rights_issues = Column(Float)
    rights_raising_fund = Column(Float)


class DividendDetail(DividendFinancingBase, Mixin):
    __tablename__ = "dividend_detail"

    provider = Column(String(length=32))
    code = Column(String(length=32))

    # 公告日
    announce_date = Column(DateTime)
    # 股权登记日
    record_date = Column(DateTime)
    # 除权除息日
    dividend_date = Column(DateTime)

    # 方案
    dividend = Column(String(length=128))


class SpoDetail(DividendFinancingBase, Mixin):
    __tablename__ = "spo_detail"

    provider = Column(String(length=32))
    code = Column(String(length=32))

    spo_issues = Column(Float)
    spo_price = Column(Float)
    spo_raising_fund = Column(Float)


class RightsIssueDetail(DividendFinancingBase, Mixin):
    __tablename__ = "rights_issue_detail"

    provider = Column(String(length=32))
    code = Column(String(length=32))

    # 配股
    rights_issues = Column(Float)
    rights_issue_price = Column(Float)
    rights_raising_fund = Column(Float)


register_schema(providers=['eastmoney'], db_name='dividend_financing', schema_base=DividendFinancingBase, entity_type='stock')

# the __all__ is generated
__all__ = ['DividendFinancing', 'DividendDetail', 'SpoDetail', 'RightsIssueDetail']


# -*- coding: utf-8 -*-

from sqlalchemy import Column, String, DateTime, BigInteger, Float
from sqlalchemy.orm import declarative_base

from zvtm.contract import TradableEntity
from zvtm.contract.register import register_schema, register_entity

StockMetaBase = declarative_base()


# 个股
@register_entity(entity_type='stock')
class Stock(StockMetaBase, TradableEntity):
    __tablename__ = 'stock'


# 个股详情
class StockDetail(StockMetaBase, TradableEntity):
    __tablename__ = 'stock_detail'

    # 所属行业
    industries = Column(String(length=128))
    # 行业指数
    industry_indices = Column(String(length=128))
    # 所属板块
    concept_indices = Column(String(length=128))
    # 所属区域
    area_indices = Column(String(length=128))

    # 成立日期
    date_of_establishment = Column(DateTime)
    # 公司简介
    profile = Column(String(length=4096))
    # 主营业务
    main_business = Column(String(length=512))
    # 发行量(股)
    issues = Column(BigInteger)
    # 发行价格
    price = Column(Float)
    # 募资净额(元)
    raising_fund = Column(Float)
    # 发行市盈率
    issue_pe = Column(Float)
    # 网上中签率
    net_winning_rate = Column(Float)


register_schema(providers=['exchange', 'joinquant', 'eastmoney', 'em'], db_name='stock_meta',
                schema_base=StockMetaBase)
# the __all__ is generated
__all__ = ['Stock', 'StockDetail']



import operator
from enum import Enum
from itertools import accumulate
from typing import List

import pandas as pd
from pandas import DataFrame

from zvtm.contract import IntervalLevel
from zvtm.contract.drawer import Drawer
from zvtm.contract.factor import FilterFactor, ScoreFactor, Factor
from zvtm.domain.meta.stock_meta import Stock
from zvtm.utils.pd_utils import index_df, pd_is_not_null
from zvtm.utils.time_utils import to_pd_timestamp, now_pd_timestamp


class TargetType(Enum):
    # open_long 代表开多，并应该平掉相应标的的空单
    open_long = 'open_long'
    # open_short 代表开空，并应该平掉相应标的的多单
    open_short = 'open_short'
    # 其他情况就是保持当前的持仓


class SelectMode(Enum):
    condition_and = 'condition_and'
    condition_or = 'condition_or'


class TargetSelector(object):
    def __init__(self, entity_ids=None, entity_schema=Stock, exchanges=None, codes=None, start_timestamp=None,
                 end_timestamp=None, long_threshold=0.8, short_threshold=0.2, level=IntervalLevel.LEVEL_1DAY,
                 provider=None, select_mode: SelectMode = SelectMode.condition_and) -> None:
        self.entity_ids = entity_ids
        self.entity_schema = entity_schema
        self.exchanges = exchanges
        self.codes = codes
        self.provider = provider
        self.select_mode = select_mode

        if start_timestamp:
            self.start_timestamp = to_pd_timestamp(start_timestamp)
        if end_timestamp:
            self.end_timestamp = to_pd_timestamp(end_timestamp)
        else:
            self.end_timestamp = now_pd_timestamp()

        self.long_threshold = long_threshold
        self.short_threshold = short_threshold
        self.level = level

        self.filter_factors: List[FilterFactor] = []
        self.score_factors: List[ScoreFactor] = []
        self.filter_result = None
        self.score_result = None

        self.open_long_df: DataFrame = None
        self.open_short_df: DataFrame = None

        self.init_factors(entity_ids=entity_ids, entity_schema=entity_schema, exchanges=exchanges, codes=codes,
                          start_timestamp=start_timestamp, end_timestamp=end_timestamp, level=self.level)

    def init_factors(self, entity_ids, entity_schema, exchanges, codes, start_timestamp, end_timestamp, level):
        pass

    def add_filter_factor(self, factor: FilterFactor):
        self.check_factor(factor)
        self.filter_factors.append(factor)
        return self

    def add_score_factor(self, factor: ScoreFactor):
        self.check_factor(factor)
        self.score_factors.append(factor)
        return self

    def check_factor(self, factor: Factor):
        assert factor.level == self.level

    def move_on(self, to_timestamp=None, kdata_use_begin_time=False, timeout=20):
        if self.score_factors:
            for factor in self.score_factors:
                factor.move_on(to_timestamp, timeout=timeout)
        if self.filter_factors:
            for factor in self.filter_factors:
                factor.move_on(to_timestamp, timeout=timeout)

        self.run()

    def run(self):
        """

        """
        if self.filter_factors:
            musts = []
            for factor in self.filter_factors:
                df = factor.result_df

                if not pd_is_not_null(df):
                    raise Exception('no data for factor:{},{}'.format(factor.factor_name, factor))

                if len(df.columns) > 1:
                    s = df.agg("and", axis="columns")
                    s.name = 'score'
                    musts.append(s.to_frame(name='score'))
                else:
                    df.columns = ['score']
                    musts.append(df)

            if self.select_mode == SelectMode.condition_and:
                self.filter_result = list(accumulate(musts, func=operator.__and__))[-1]
            else:
                self.filter_result = list(accumulate(musts, func=operator.__or__))[-1]
        if self.score_factors:
            scores = []
            for factor in self.score_factors:
                df = factor.result_df
                if not pd_is_not_null(df):
                    raise Exception('no data for factor:{},{}'.format(factor.factor_name, factor))

                if len(df.columns) > 1:
                    s = df.agg("mean", axis="columns")
                    s.name = 'score'
                    scores.append(s.to_frame(name='score'))
                else:
                    df.columns = ['score']
                    scores.append(df)
            self.score_result = list(accumulate(scores, func=operator.__add__))[-1]

        self.generate_targets()

    def get_targets(self, timestamp, target_type: TargetType = TargetType.open_long) -> pd.DataFrame:
        if target_type == TargetType.open_long:
            df = self.open_long_df
        if target_type == TargetType.open_short:
            df = self.open_short_df

        if pd_is_not_null(df):
            if timestamp in df.index:
                target_df = df.loc[[to_pd_timestamp(timestamp)], :]
                return target_df['entity_id'].tolist()
        return []

    def get_open_long_targets(self, timestamp):
        return self.get_targets(timestamp=timestamp, target_type=TargetType.open_long)

    def get_open_short_targets(self, timestamp):
        return self.get_targets(timestamp=timestamp, target_type=TargetType.open_short)

    # overwrite it to generate targets
    def generate_targets(self):
        if pd_is_not_null(self.filter_result) and pd_is_not_null(self.score_result):
            # for long
            result1 = self.filter_result[self.filter_result.score]
            result2 = self.score_result[self.score_result.score >= self.long_threshold]
            long_result = result2.loc[result1.index, :]
            # for short
            result1 = self.filter_result[~self.filter_result.score]
            result2 = self.score_result[self.score_result.score <= self.short_threshold]
            short_result = result2.loc[result1.index, :]
        elif pd_is_not_null(self.score_result):
            long_result = self.score_result[self.score_result.score >= self.long_threshold]
            short_result = self.score_result[self.score_result.score <= self.short_threshold]
        else:
            long_result = self.filter_result[self.filter_result.score == True]
            short_result = self.filter_result[self.filter_result.score == False]

        self.open_long_df = self.normalize_result_df(long_result)
        self.open_short_df = self.normalize_result_df(short_result)

    def get_result_df(self):
        return self.open_long_df

    def normalize_result_df(self, df):
        if pd_is_not_null(df):
            df = df.reset_index()
            df = index_df(df)
            df = df.sort_values(by=['score', 'entity_id'])
        return df

    def draw(self,
             render='html',
             file_name=None,
             width=None,
             height=None,
             title=None,
             keep_ui_state=True,
             annotation_df=None,
             target_type: TargetType = TargetType.open_long):

        if target_type == TargetType.open_long:
            df = self.open_long_df.copy()
        elif target_type == TargetType.open_short:
            df = self.open_short_df.copy()

        df['target_type'] = target_type.value

        if pd_is_not_null(df):
            df = df.reset_index(drop=False)
            drawer = Drawer(df)

            drawer.draw_table(width=width, height=height, title=title,
                              keep_ui_state=keep_ui_state)


# the __all__ is generated
__all__ = ['TargetType', 'TargetSelector']



# -*- coding: utf-8 -*-
import pandas as pd
from jqdatapy.api import get_all_securities, run_query

from zvtm.api.portfolio import portfolio_relate_stock
from zvtm.api.utils import china_stock_code_to_id
from zvtm.contract.api import df_to_db, get_entity_exchange, get_entity_code
from zvtm.contract.recorder import Recorder, TimeSeriesDataRecorder
from zvtm.domain import EtfStock, Stock, Etf, StockDetail,Index,IndexStock
from zvtm.recorders.joinquant.common import to_entity_id, jq_to_report_period
from zvtm.utils.pd_utils import pd_is_not_null
from zvtm.utils.time_utils import to_time_str

from jqdatasdk import get_index_stocks,auth,get_index_weights
from zvtm import zvt_config

class BaseJqChinaMetaRecorder(Recorder):
    provider = 'joinquant'

    def __init__(self, force_update=True, sleeping_time=10) -> None:
        super().__init__(force_update, sleeping_time)

    def to_zvt_entity(self, df, entity_type, category=None):
        df = df.set_index('code')
        df.index.name = 'entity_id'
        df = df.reset_index()
        # 上市日期
        df.rename(columns={'start_date': 'timestamp'}, inplace=True)
        df['timestamp'] = pd.to_datetime(df['timestamp'])
        df['list_date'] = df['timestamp']
        df['end_date'] = pd.to_datetime(df['end_date'])

        df['entity_id'] = df['entity_id'].apply(lambda x: to_entity_id(entity_type=entity_type, jq_code=x))
        df['id'] = df['entity_id']
        df['entity_type'] = entity_type
        df['exchange'] = df['entity_id'].apply(lambda x: get_entity_exchange(x))
        df['code'] = df['entity_id'].apply(lambda x: get_entity_code(x))
        df['name'] = df['display_name']

        if category:
            df['category'] = category

        return df


class JqChinaStockRecorder(BaseJqChinaMetaRecorder):
    data_schema = Stock

    def run(self):
        # 抓取股票列表
        df_stock = self.to_zvt_entity(get_all_securities(code='stock'), entity_type='stock')
        df_to_db(df_stock, data_schema=Stock, provider=self.provider, force_update=self.force_update)
        # persist StockDetail too
        df_to_db(df=df_stock, data_schema=StockDetail, provider=self.provider, force_update=self.force_update)

        # self.logger.info(df_stock)
        self.logger.info("persist stock list success")


class JqChinaEtfRecorder(BaseJqChinaMetaRecorder):
    data_schema = Etf

    def run(self):
        # 抓取etf列表
        df_index = self.to_zvt_entity(get_all_securities(code='etf'), entity_type='etf', category='etf')
        df_to_db(df_index, data_schema=Etf, provider=self.provider, force_update=self.force_update)

        # self.logger.info(df_index)
        self.logger.info("persist etf list success")


class JqChinaStockEtfPortfolioRecorder(TimeSeriesDataRecorder):
    entity_provider = 'joinquant'
    entity_schema = Etf

    # 数据来自jq
    provider = 'joinquant'

    data_schema = EtfStock

    def record(self, entity, start, end, size, timestamps):
        df = run_query(table='finance.FUND_PORTFOLIO_STOCK',
                       conditions=f'pub_date#>=#{to_time_str(start)}&code#=#{entity.code}',
                       parse_dates=None)
        if pd_is_not_null(df):
            #          id    code period_start  period_end    pub_date  report_type_id report_type  rank  symbol  name      shares    market_cap  proportion
            # 0   8640569  159919   2018-07-01  2018-09-30  2018-10-26          403003        第三季度     1  601318  中国平安  19869239.0  1.361043e+09        7.09
            # 1   8640570  159919   2018-07-01  2018-09-30  2018-10-26          403003        第三季度     2  600519  贵州茅台    921670.0  6.728191e+08        3.50
            # 2   8640571  159919   2018-07-01  2018-09-30  2018-10-26          403003        第三季度     3  600036  招商银行  18918815.0  5.806184e+08        3.02
            # 3   8640572  159919   2018-07-01  2018-09-30  2018-10-26          403003        第三季度     4  601166  兴业银行  22862332.0  3.646542e+08        1.90
            df['timestamp'] = pd.to_datetime(df['pub_date'])

            df.rename(columns={'symbol': 'stock_code', 'name': 'stock_name'}, inplace=True)
            df['proportion'] = df['proportion'] * 0.01

            df = portfolio_relate_stock(df, entity)

            df['stock_id'] = df['stock_code'].apply(lambda x: china_stock_code_to_id(x))
            df['id'] = df[['entity_id', 'stock_id', 'pub_date', 'id']].apply(lambda x: '_'.join(x.astype(str)), axis=1)
            df['report_date'] = pd.to_datetime(df['period_end'])
            df['report_period'] = df['report_type'].apply(lambda x: jq_to_report_period(x))

            df_to_db(df=df, data_schema=self.data_schema, provider=self.provider, force_update=self.force_update)

            # self.logger.info(df.tail())
            self.logger.info(f"persist etf {entity.code} portfolio success {df.iloc[-1]['pub_date']}")

        return None

class JqChinaIndexRecorder(BaseJqChinaMetaRecorder):
    data_schema = Index

    def run(self):
        # 抓取指数列表
        df_index = self.to_zvt_entity(get_all_securities(code='index'), entity_type='index', category='index')
        df_to_db(df_index, data_schema=Index, provider=self.provider, force_update=self.force_update)
        entity_type='index-stock'
        category='index_stock'
        for index in range(len(df_index['code'])):
            df_index_stock_ins = pd.DataFrame(columns=["id", "entity_id", "timestamp", "entity_type",'exchange','code','name','stock_id','stock_code','stock_name','report_period','report_date','proportion','shares','market_cap'])
            jqstock = df_index['code'][index]+'.XSHG'
            auth(zvt_config['jq_username'], zvt_config['jq_password'])
            df_index_stock = get_index_stocks(jqstock)
            if len(df_index_stock)>0:
                for istock in df_index_stock:
                    df =  Stock.query_data(provider='joinquant',code=istock[0:6])
                    df_index_stock_weight = get_index_weights(index_id=istock)
                    if pd_is_not_null(df):
                        stock_code = '-' + istock[0:6]
                        df['stock_id'] = df['entity_id']
                        df['entity_id'] = entity_type + df_index['entity_id'][index][5:] + stock_code   #   [str(x) + stock_code for x in df_index['entity_id'][index]]
                        df['id'] = df['entity_id']
                        df['entity_type'] = entity_type

                        df['stock_code'] = df['code']
                        df['stock_name'] = df['name']

                        df['code'] = istock[0:6]
                        df['name'] = df_index['name'][index]
                        #指数股票权重
                        df['proportion'] =df_index_stock_weight['weight']
                        df['report_date'] = df_index_stock_weight['date']
                        df['report_period'] = ''
                        df['shares'] = ''
                        df['market_cap'] = ''
                        #df_index_stock_ins.append(df, ignore_index=True)

                        df_index_stock_ins = pd.merge(df_index_stock_ins,df,how='left',on=["id", "entity_id", "timestamp", "entity_type",'exchange','code','name','stock_id','stock_code','stock_name','report_period','report_date','proportion','shares','market_cap'])




                df_to_db(df_index_stock_ins, data_schema=IndexStock, provider=self.provider, force_update=self.force_update)
        self.logger.info("persist index and indexstock list success")



if __name__ == '__main__':
    #JqChinaEtfRecorder().run()
    JqChinaIndexRecorder().run()

    #JqChinaStockEtfPortfolioRecorder(codes=['510050']).run()
# the __all__ is generated
__all__ = ['BaseJqChinaMetaRecorder', 'JqChinaStockRecorder', 'JqChinaEtfRecorder', 'JqChinaStockEtfPortfolioRecorder']


# -*- coding: utf-8 -*-

CHINA_STOCK_MAIN_INDEX = [
    # # 聚宽编码
    # # 市场通编码	市场通名称
    # # 310001	沪股通
    # # 310002	深股通
    # # 310003	港股通（沪）
    # # 310004	港股通（深）
    {'id': 'index_cn_310001',
     'entity_id': 'index_cn_310001',
     'code': '310001',
     'name': '沪股通',
     'timestamp': '2014-11-17',
     'exchange': 'cn',
     'entity_type': 'index',
     'category': 'other'},
    {'id': 'index_cn_310002',
     'entity_id': 'index_cn_310002',
     'code': '310002',
     'name': '深股通',
     'timestamp': '2014-11-17',
     'exchange': 'cn',
     'entity_type': 'index',
     'category': 'other'},
    {'id': 'index_cn_310003',
     'entity_id': 'index_cn_310003',
     'code': '310003',
     'name': '港股通（沪）',
     'timestamp': '2014-11-17',
     'exchange': 'cn',
     'entity_type': 'index',
     'category': 'other'},
    {'id': 'index_cn_310004',
     'entity_id': 'index_cn_310004',
     'code': '310004',
     'name': '港股通（深）',
     'timestamp': '2014-11-17',
     'exchange': 'cn',
     'entity_type': 'index',
     'category': 'other'}
]


def init_main_index(provider='exchange'):
    from zvtm.utils.time_utils import to_pd_timestamp
    import pandas as pd
    from zvtm.contract.api import df_to_db
    from zvtm.domain import Index

    for item in CHINA_STOCK_MAIN_INDEX:
        item['timestamp'] = to_pd_timestamp(item['timestamp'])
    df = pd.DataFrame(CHINA_STOCK_MAIN_INDEX)
    # print(df)
    df_to_db(df=df, data_schema=Index, provider=provider, force_update=False)


init_main_index(provider='exchange')

# the __all__ is generated
__all__ = ['init_main_index']

# __init__.py structure:
# common code of the package
# export interface in __all__ which contains __all__ of its sub modules

# import all from submodule joinquant
from .joinquant import *
from .joinquant import __all__ as _joinquant_all
__all__ += _joinquant_all

# import all from submodule exchange
from .exchange import *
from .exchange import __all__ as _exchange_all
__all__ += _exchange_all

# import all from submodule em
from .em import *
from .em import __all__ as _em_all
__all__ += _em_all

# import all from submodule consts
from .consts import *
from .consts import __all__ as _consts_all
__all__ += _consts_all

# import all from submodule eastmoney
from .eastmoney import *
from .eastmoney import __all__ as _eastmoney_all
__all__ += _eastmoney_all

# import all from submodule sina
from .sina import *
from .sina import __all__ as _sina_all
__all__ += _sina_all


# -*- coding: utf-8 -*-
import os
from typing import List


def list_all_files(dir_path: str = './domain', ext: str = '.py', excludes=None, return_base_name=False) -> List[str]:
    """
    list all files with extension in specific directory recursively

    :param dir_path: the directory path
    :param ext: file extension
    :param excludes: excluding files
    :param return_base_name: return file name if True otherwise abs path
    :return:
    """
    files = []
    for entry in os.scandir(dir_path):
        if entry.is_dir():
            files += list_all_files(entry.path)
        elif entry.is_file():
            if not ext or (ext and entry.path.endswith(ext)):
                if excludes and entry.path.endswith(excludes):
                    continue
                if return_base_name:
                    files.append(os.path.basename(entry.path))
                else:
                    files.append(entry.path)
        else:
            pass
    return files
# the __all__ is generated
__all__ = ['list_all_files']


# -*- coding: utf-8 -*-
import logging
import os
from typing import List

from zvtm.autocode.templates import all_tpls
from zvtm.contract import IntervalLevel, AdjustType
from zvtm.utils import now_pd_timestamp
from zvtm.utils.file_utils import list_all_files
from zvtm.utils.git_utils import get_git_user_name, get_git_user_email

logger = logging.getLogger(__name__)


def all_sub_modules(dir_path: str):
    """
    list all module name in specific directory

    :param dir_path:
    :return:
    """
    modules = []
    for entry in os.scandir(dir_path):
        if entry.is_dir() or (
                entry.path.endswith('.py') and not entry.path.endswith('__init__.py')):
            module_name = os.path.splitext(os.path.basename(entry.path))[0]
            # ignore hidden
            if module_name.startswith('.') or not module_name[0].isalpha():
                continue
            modules.append(module_name)
    return modules


def _remove_start_end(line: str, start='class ', end='('):
    if line.startswith(start) and (end in line):
        start_index = len(start)
        end_index = line.index(end)
        return line[start_index:end_index]
    if not start and (end in line):
        end_index = line.index(end)
        return line[:end_index]


def _get_interface_name(line: str):
    """
    get interface name of the line

    :param line: the line of the source
    :return:
    """
    if line.startswith("class "):
        return _remove_start_end(line, "class ", "(")
    elif line.startswith("def "):
        return _remove_start_end(line, "def ", "(")


def all_sub_all(sub_module):
    return '''

# import all from submodule {0}
from .{0} import *
from .{0} import __all__ as _{0}_all
__all__ += _{0}_all'''.format(sub_module)


def fill_package_if_not_exist(dir_path: str):
    fill_package(dir_path)
    for entry in os.scandir(dir_path):
        if entry.is_dir():
            fill_package(entry.path)
            fill_package_if_not_exist(entry.path)
        elif entry.is_file():
            pass


def fill_package(dir_path: str):
    base_name = os.path.basename(dir_path)
    if base_name[0].isalpha():
        if os.path.isdir(dir_path):
            pkg_file = os.path.join(dir_path, '__init__.py')
            if not os.path.exists(pkg_file):
                package_template = '# -*- coding: utf-8 -*-\n'
                with open(pkg_file, 'w') as outfile:
                    outfile.write(package_template)


def gen_exports(dir_path='./domain', gen_flag='# the __all__ is generated', export_from_package=True):
    fill_package_if_not_exist(dir_path=dir_path)
    files = list_all_files(dir_path=dir_path)
    for file in files:
        exports = []
        lines = []
        # read and generate __all__
        with open(file) as fp:
            line = fp.readline()
            while line:
                if line.startswith(gen_flag):
                    break
                lines.append(line)
                export = _get_interface_name(line)
                if export and export[0].isalpha():
                    exports.append(export)
                line = fp.readline()
        print(f'{file}:{exports}')
        lines.append(gen_flag)
        lines.append(f'\n__all__ = {exports}')

        # the package module
        if export_from_package:
            basename = os.path.basename(file)
            if basename == '__init__.py':
                dir_path = os.path.dirname(file)
                modules = all_sub_modules(dir_path)
                if modules:
                    lines.append('''

# __init__.py structure:
# common code of the package
# export interface in __all__ which contains __all__ of its sub modules''')
                    for mod in modules:
                        lines.append(all_sub_all(mod))

        # write with __all__
        with open(file, mode='w') as fp:
            fp.writelines(lines)


# kdata schema rule
# 1)name:{entity_type.capitalize()}{IntervalLevel.value.capitalize()}Kdata
# 2)one db file for one schema

def gen_kdata_schema(pkg: str,
                     providers: List[str],
                     entity_type: str,
                     levels: List[IntervalLevel],
                     adjust_types=None,
                     entity_in_submodule: bool = False,
                     kdata_module='quotes'):
    if adjust_types is None:
        adjust_types = [None]
    tables = []

    base_path = './domain'

    if kdata_module:
        base_path = os.path.join(base_path, kdata_module)
    if entity_in_submodule:
        base_path = os.path.join(base_path, entity_type)

    if not os.path.exists(base_path):
        logger.info(f'create dir {base_path}')
        os.makedirs(base_path)

    for level in levels:

        for adjust_type in adjust_types:
            level = IntervalLevel(level)

            cap_entity_type = entity_type.capitalize()
            cap_level = level.value.capitalize()

            # you should define {EntityType}KdataCommon in kdata_module at first
            kdata_common = f'{cap_entity_type}KdataCommon'

            if adjust_type and (adjust_type != AdjustType.qfq):
                class_name = f'{cap_entity_type}{cap_level}{adjust_type.value.capitalize()}Kdata'
                table_name = f'{entity_type}_{level.value}_{adjust_type.value.lower()}_kdata'

            else:
                class_name = f'{cap_entity_type}{cap_level}Kdata'
                table_name = f'{entity_type}_{level.value}_kdata'

            tables.append(table_name)

            schema_template = f'''# -*- coding: utf-8 -*-
# this file is generated by gen_kdata_schema function, dont't change it
from sqlalchemy.orm import declarative_base

from zvtm.contract.register import register_schema
from {pkg}.domain.{kdata_module} import {kdata_common}

KdataBase = declarative_base()


class {class_name}(KdataBase, {kdata_common}):
    __tablename__ = '{table_name}'


register_schema(providers={providers}, db_name='{table_name}', schema_base=KdataBase, entity_type='{entity_type}')

'''
            # generate the schema
            with open(os.path.join(base_path, f'{table_name}.py'), 'w') as outfile:
                outfile.write(schema_template)

        # generate the package
        pkg_file = os.path.join(base_path, '__init__.py')
        if not os.path.exists(pkg_file):
            package_template = '''# -*- coding: utf-8 -*-
'''
            with open(pkg_file, 'w') as outfile:
                outfile.write(package_template)

    # generate exports
    gen_exports('./domain')


def gen_plugin_project(entity_type, prefix: str = 'zvt', dir_path: str = '.', providers=['joinquant']):
    """
    generate a standard plugin project

    :param entity_type: the entity type of the plugin project
    :param prefix:  project prefix
    :param dir_path: the root path for the project
    :param providers: the supported providers
    """

    # generate project files
    project = f'{prefix}_{entity_type}'
    entity_class = entity_type.capitalize()
    project_path = os.path.join(dir_path, project)
    if not os.path.exists(project_path):
        os.makedirs(project_path)

    current_time = now_pd_timestamp()
    user_name = get_git_user_name()
    user_email = get_git_user_email()

    for tpl in all_tpls(project=project, entity_type=entity_type):
        file_name = tpl[0]
        tpl_content = tpl[1].safe_substitute(project=project,
                                             entity_type=entity_type,
                                             entity_class=entity_class,
                                             providers=providers,
                                             provider=providers[0],
                                             Provider=providers[0].capitalize(),
                                             year=current_time.year,
                                             user=user_name,
                                             email=user_email)
        file_path = os.path.join(project_path, file_name)

        file_dir = os.path.dirname(file_path)
        if not os.path.exists(file_dir):
            os.makedirs(file_dir)

        with open(file_path, "w", encoding="utf-8") as fh:
            fh.write(tpl_content)


# the __all__ is generated
__all__ = ['all_sub_modules', 'all_sub_all', 'fill_package_if_not_exist', 'fill_package', 'gen_exports',
           'gen_kdata_schema', 'gen_plugin_project']



# -*- coding: utf-8 -*-
from zvtm.api.utils import to_report_period_type
from zvtm.domain import CashFlowStatement
from zvtm.recorders.eastmoney.finance.base_china_stock_finance_recorder import BaseChinaStockFinanceRecorder
from zvtm.utils.time_utils import to_pd_timestamp
from zvtm.utils.utils import add_func_to_value, first_item_to_float

cash_flow_map = {
    # 经营活动产生的现金流量
    #
    # 销售商品、提供劳务收到的现金
    "cash_from_selling": "Salegoodsservicerec",

    # 收到的税费返还
    "tax_refund": "Taxreturnrec",

    # 收到其他与经营活动有关的现金
    "cash_from_other_op": "Otheroperaterec",

    # 经营活动现金流入小计
    "total_op_cash_inflows": "Sumoperateflowin",

    # 购买商品、接受劳务支付的现金
    "cash_to_goods_services": "Buygoodsservicepay",
    # 支付给职工以及为职工支付的现金
    "cash_to_employees": "Employeepay",
    # 支付的各项税费
    "taxes_and_surcharges": "Taxpay",
    # 支付其他与经营活动有关的现金
    "cash_to_other_related_op": "Otheroperatepay",
    # 经营活动现金流出小计
    "total_op_cash_outflows": "Sumoperateflowout",

    # 经营活动产生的现金流量净额
    "net_op_cash_flows": "Netoperatecashflow",

    # 投资活动产生的现金流量

    # 收回投资收到的现金
    "cash_from_disposal_of_investments": "Disposalinvrec",
    # 取得投资收益收到的现金
    "cash_from_returns_on_investments": "Invincomerec",
    # 处置固定资产、无形资产和其他长期资产收回的现金净额
    "cash_from_disposal_fixed_intangible_assets": "Dispfilassetrec",
    # 处置子公司及其他营业单位收到的现金净额
    "cash_from_disposal_subsidiaries": "Dispsubsidiaryrec",

    # 收到其他与投资活动有关的现金
    "cash_from_other_investing": "Otherinvrec",

    # 投资活动现金流入小计
    "total_investing_cash_inflows": "Suminvflowin",

    # 购建固定资产、无形资产和其他长期资产支付的现金
    "cash_to_acquire_fixed_intangible_assets": "Buyfilassetpay",
    # 投资支付的现金
    "cash_to_investments": "Invpay",

    # 取得子公司及其他营业单位支付的现金净额
    "cash_to_acquire_subsidiaries": "Getsubsidiarypay",

    # 支付其他与投资活动有关的现金
    "cash_to_other_investing": "Otherinvpay",

    # 投资活动现金流出小计
    "total_investing_cash_outflows": "Suminvflowout",

    # 投资活动产生的现金流量净额
    "net_investing_cash_flows": "Netinvcashflow",

    # 筹资活动产生的现金流量
    #
    # 吸收投资收到的现金
    "cash_from_accepting_investment": "Acceptinvrec",
    # 子公司吸收少数股东投资收到的现金
    "cash_from_subsidiaries_accepting_minority_interest": "Subsidiaryaccept",

    # 取得借款收到的现金
    "cash_from_borrowings": "Loanrec",
    # 发行债券收到的现金
    "cash_from_issuing_bonds": "Issuebondrec",
    # 收到其他与筹资活动有关的现金
    "cash_from_other_financing": "Otherfinarec",

    # 筹资活动现金流入小计
    "total_financing_cash_inflows": "Sumfinaflowin",

    # 偿还债务支付的现金
    "cash_to_repay_borrowings": "Repaydebtpay",

    # 分配股利、利润或偿付利息支付的现金
    "cash_to_pay_interest_dividend": "Diviprofitorintpay",

    # 子公司支付给少数股东的股利、利润
    "cash_to_pay_subsidiaries_minority_interest": "Subsidiarypay",

    # 支付其他与筹资活动有关的现金
    "cash_to_other_financing": "Otherfinapay",
    # 筹资活动现金流出小计
    "total_financing_cash_outflows": "Sumfinaflowout",

    # 筹资活动产生的现金流量净额
    "net_financing_cash_flows": "Netfinacashflow",
    # 汇率变动对现金及现金等价物的影响
    "foreign_exchange_rate_effect": "Effectexchangerate",
    # 现金及现金等价物净增加额
    "net_cash_increase": "Nicashequi",
    # 加: 期初现金及现金等价物余额
    "cash_at_beginning": "Cashequibeginning",
    # 期末现金及现金等价物余额
    "cash": "Cashequiending",

    # 银行相关
    # 客户存款和同业及其他金融机构存放款项净增加额
    "fi_deposit_increase": "Nideposit",
    # 向中央银行借款净增加额
    "fi_borrow_from_central_bank_increase": "Niborrowfromcbank",
    # 存放中央银行和同业款项及其他金融机构净减少额
    "fi_deposit_in_others_decrease": "Nddepositincbankfi",
    # 拆入资金及卖出回购金融资产款净增加额
    "fi_borrowing_and_sell_repurchase_increase": "Niborrowsellbuyback",
    # 其中:卖出回购金融资产款净增加额
    "fi_sell_repurchase_increase": "Nisellbuybackfasset",
    # 拆出资金及买入返售金融资产净减少额
    "fi_lending_and_buy_repurchase_decrease": "Ndlendbuysellback",
    # 其中:拆出资金净减少额
    "fi_lending_decrease": "Ndlendfund",
    # 买入返售金融资产净减少额
    "fi_buy_repurchase_decrease": "Ndbuysellbackfasset",
    # 收取的利息、手续费及佣金的现金
    "fi_cash_from_interest_commission": "Intandcommrec",
    # 客户贷款及垫款净增加额
    "fi_loan_advance_increase": "Niloanadvances",
    # 存放中央银行和同业及其他金融机构款项净增加额
    "fi_deposit_in_others_increase": "Nidepositincbankfi",
    # 拆出资金及买入返售金融资产净增加额
    "fi_lending_and_buy_repurchase_increase": "Nilendsellbuyback",
    # 其中:拆出资金净增加额
    "fi_lending_increase": "Nilendfund",
    # 拆入资金及卖出回购金融资产款净减少额
    "fi_borrowing_and_sell_repurchase_decrease": "Ndborrowsellbuyback",
    # 其中:拆入资金净减少额
    "fi_borrowing_decrease": "Ndborrowfund",
    # 卖出回购金融资产净减少额
    "fi_sell_repurchase_decrease": "Ndsellbuybackfasset",
    # 支付利息、手续费及佣金的现金
    "fi_cash_to_interest_commission": "Intandcommpay",
    # 应收账款净增加额
    "fi_account_receivable_increase": "Niaccountrec",
    # 偿付债券利息支付的现金
    "fi_cash_to_pay_interest": "Bondintpay",

    # 保险相关
    # 收到原保险合同保费取得的现金
    "fi_cash_from_premium_of_original": "Premiumrec",
    # 保户储金及投资款净增加额
    "fi_insured_deposit_increase": "Niinsureddepositinv",
    # 银行及证券业务卖出回购资金净增加额
    "fi_bank_broker_sell_repurchase_increase": "Nisellbuyback",
    # 银行及证券业务买入返售资金净减少额
    "fi_bank_broker_buy_repurchase_decrease": "Ndbuysellback",
    # 支付原保险合同赔付等款项的现金
    "fi_cash_to_insurance_claim": "Indemnitypay",
    # 支付再保险业务现金净额
    "fi_cash_to_reinsurance": "Netripay",
    # 银行业务及证券业务拆借资金净减少额
    "fi_lending_decrease": "Ndlendfund",
    # 银行业务及证券业务卖出回购资金净减少额
    "fi_bank_broker_sell_repurchase_decrease": "Ndsellbuyback",
    # 支付保单红利的现金
    "fi_cash_to_dividends": "Divipay",
    # 保户质押贷款净增加额
    "fi_insured_pledge_loans_increase": "Niinsuredpledgeloan",
    # 收购子公司及其他营业单位支付的现金净额
    "fi_cash_to_acquire_subsidiaries": "Buysubsidiarypay",
    # 处置子公司及其他营业单位流出的现金净额
    "fi_cash_to_disposal_subsidiaries": "Dispsubsidiarypay",
    # 支付卖出回购金融资产款现金净额
    "fi_cash_to_sell_repurchase": "Netsellbuybackfassetpay",

    # 券商相关
    # 拆入资金净增加额
    "fi_borrowing_increase": "Niborrowfund",
    # 代理买卖证券收到的现金净额
    "fi_cash_from_trading_agent": "Agenttradesecurityrec",
    # 回购业务资金净增加额
    "fi_cash_from_repurchase_increase": "Nibuybackfund",
    # 处置交易性金融资产的净减少额
    "fi_disposal_trade_asset_decrease": "Nddisptradefasset",
    # 回购业务资金净减少额
    "fi_repurchase_decrease": "Ndbuybackfund",
    # 代理买卖证券支付的现金净额（净减少额）
    "fi_cash_to_agent_trade": "Agenttradesecuritypay",
}

add_func_to_value(cash_flow_map, first_item_to_float)
cash_flow_map["report_period"] = ("ReportDate", to_report_period_type)
cash_flow_map["report_date"] = ("ReportDate", to_pd_timestamp)


class ChinaStockCashFlowRecorder(BaseChinaStockFinanceRecorder):
    data_schema = CashFlowStatement

    url = 'https://emh5.eastmoney.com/api/CaiWuFenXi/GetXianJinLiuLiangBiaoList'
    finance_report_type = 'XianJinLiuLiangBiaoList'
    data_type = 4

    def get_data_map(self):
        return cash_flow_map


if __name__ == '__main__':
    # init_log('cash_flow.log')
    recorder = ChinaStockCashFlowRecorder(codes=['002572'])
    recorder.run()
# the __all__ is generated
__all__ = ['ChinaStockCashFlowRecorder']


# -*- coding: utf-8 -*-
import inspect
from datetime import timedelta
from typing import List, Union

import pandas as pd
from sqlalchemy import Column, String, DateTime, Float
from sqlalchemy.orm import Session

from zvtm.contract import IntervalLevel
from zvtm.utils.time_utils import date_and_time, is_same_time, now_pd_timestamp


class Mixin(object):
    id = Column(String(length=128), primary_key=True)
    # entity id for this model
    entity_id = Column(String(length=128))

    # the meaning could be different for different case,most of time it means 'happen time'
    timestamp = Column(DateTime)

    # unix epoch,same meaning with timestamp
    # ts = Column(BIGINT)

    @classmethod
    def help(cls):
        print(inspect.getsource(cls))

    @classmethod
    def important_cols(cls):
        return []

    @classmethod
    def time_field(cls):
        return 'timestamp'

    @classmethod
    def register_recorder_cls(cls, provider, recorder_cls):
        """
        register the recorder for the schema

        :param provider:
        :param recorder_cls:
        """
        # don't make provider_map_recorder as class field,it should be created for the sub class as need
        if not hasattr(cls, 'provider_map_recorder'):
            cls.provider_map_recorder = {}

        if provider not in cls.provider_map_recorder:
            cls.provider_map_recorder[provider] = recorder_cls

    @classmethod
    def register_provider(cls, provider):
        # don't make providers as class field,it should be created for the sub class as need
        if not hasattr(cls, 'providers'):
            cls.providers = []

        if provider not in cls.providers:
            cls.providers.append(provider)

    @classmethod
    def get_providers(cls) -> List[str]:
        assert hasattr(cls, 'providers')
        return cls.providers

    @classmethod
    def test_data_correctness(cls, provider, data_samples):
        for data in data_samples:
            item = cls.query_data(provider=provider, ids=[data['id']], return_type='dict')
            print(item)
            for k in data:
                if k == 'timestamp':
                    assert is_same_time(item[0][k], data[k])
                else:
                    assert item[0][k] == data[k]

    @classmethod
    def get_one(cls, id, provider_index: int = 0, provider: str = None):
        from .api import get_one
        if not provider:
            provider = cls.providers[provider_index]
        return get_one(data_schema=cls, id=id, provider=provider)

    @classmethod
    def query_data(cls,
                   provider_index: int = 0,
                   ids: List[str] = None,
                   entity_ids: List[str] = None,
                   entity_id: str = None,
                   codes: List[str] = None,
                   code: str = None,
                   level: Union[IntervalLevel, str] = None,
                   provider: str = None,
                   columns: List = None,
                   col_label: dict = None,
                   return_type: str = 'df',
                   start_timestamp: Union[pd.Timestamp, str] = None,
                   end_timestamp: Union[pd.Timestamp, str] = None,
                   filters: List = None,
                   session: Session = None,
                   order=None,
                   limit: int = None,
                   index: Union[str, list] = None,
                   time_field: str = 'timestamp'):
        from .api import get_data
        if not provider:
            provider = cls.providers[provider_index]
        return get_data(data_schema=cls, ids=ids, entity_ids=entity_ids, entity_id=entity_id, codes=codes,
                        code=code, level=level, provider=provider, columns=columns, col_label=col_label,
                        return_type=return_type, start_timestamp=start_timestamp, end_timestamp=end_timestamp,
                        filters=filters, session=session, order=order, limit=limit, index=index, time_field=time_field)

    @classmethod
    def record_data(cls,
                    provider_index: int = 0,
                    provider: str = None,
                    force_update=None,
                    sleeping_time=None,
                    exchanges=None,
                    entity_ids=None,
                    code=None,
                    codes=None,
                    real_time=None,
                    fix_duplicate_way=None,
                    start_timestamp=None,
                    end_timestamp=None,
                    one_day_trading_minutes=None,
                    **kwargs):
        if cls.provider_map_recorder:
            print(f'{cls.__name__} registered recorders:{cls.provider_map_recorder}')

            if provider:
                recorder_class = cls.provider_map_recorder[provider]
            else:
                recorder_class = cls.provider_map_recorder[cls.providers[provider_index]]

            # get args for specific recorder class
            from zvtm.contract.recorder import TimeSeriesDataRecorder
            if issubclass(recorder_class, TimeSeriesDataRecorder):
                args = [item for item in inspect.getfullargspec(cls.record_data).args if
                        item not in ('cls', 'provider_index', 'provider')]
            else:
                args = ['force_update', 'sleeping_time']

            # just fill the None arg to kw,so we could use the recorder_class default args
            kw = {}
            for arg in args:
                tmp = eval(arg)
                if tmp is not None:
                    kw[arg] = tmp

            # FixedCycleDataRecorder
            from zvtm.contract.recorder import FixedCycleDataRecorder
            if issubclass(recorder_class, FixedCycleDataRecorder):
                # contract:
                # 1)use FixedCycleDataRecorder to record the data with IntervalLevel
                # 2)the table of schema with IntervalLevel format is {entity}_{level}_[adjust_type]_{event}
                table: str = cls.__tablename__
                try:
                    items = table.split('_')
                    if len(items) == 4:
                        adjust_type = items[2]
                        kw['adjust_type'] = adjust_type
                    level = IntervalLevel(items[1])
                except:
                    # for other schema not with normal format,but need to calculate size for remaining days
                    level = IntervalLevel.LEVEL_1DAY

                kw['level'] = level

                # add other custom args
                for k in kwargs:
                    kw[k] = kwargs[k]

                r = recorder_class(**kw)
                r.run()
                return
            else:
                r = recorder_class(**kw)
                r.run()
                return
        else:
            print(f'no recorders for {cls.__name__}')


class NormalMixin(Mixin):
    # the record created time in db
    created_timestamp = Column(DateTime, default=pd.Timestamp.now())
    # the record updated time in db, some recorder would check it for whether need to refresh
    updated_timestamp = Column(DateTime)


class Entity(Mixin):
    # 标的类型
    entity_type = Column(String(length=64))
    # 所属交易所
    exchange = Column(String(length=32))
    # 编码
    code = Column(String(length=64))
    # 名字
    name = Column(String(length=128))
    # 上市日
    list_date = Column(DateTime)
    # 退市日
    end_date = Column(DateTime)


class TradableEntity(Entity):
    @classmethod
    def get_trading_dates(cls, start_date=None, end_date=None):
        """
        overwrite it to get the trading dates of the entity

        :param start_date:
        :param end_date:
        :return:
        """
        return pd.date_range(start_date, end_date, freq='B')

    @classmethod
    def get_trading_intervals(cls):
        """
        overwrite it to get the trading intervals of the entity

        :return:[(start,end)]
        """
        return [('09:30', '11:30'), ('13:00', '15:00')]

    @classmethod
    def in_trading_time(cls, timestamp=None):
        if not timestamp:
            timestamp = now_pd_timestamp()
        else:
            timestamp = pd.Timestamp(timestamp)
        open_time = date_and_time(the_date=timestamp.date(), the_time=cls.get_trading_intervals()[0][0])
        close_time = date_and_time(the_date=timestamp.date(), the_time=cls.get_trading_intervals()[-1][1])
        return open_time < timestamp < close_time

    @classmethod
    def get_close_hour_and_minute(cls):
        hour, minute = cls.get_trading_intervals()[-1][1].split(":")
        return int(hour), int(minute)

    @classmethod
    def get_interval_timestamps(cls, start_date, end_date, level: IntervalLevel):
        """
        generate the timestamps for the level

        :param start_date:
        :param end_date:
        :param level:
        """

        for current_date in cls.get_trading_dates(start_date=start_date, end_date=end_date):
            if level == IntervalLevel.LEVEL_1DAY:
                yield current_date
            elif level == IntervalLevel.LEVEL_1WEEK:
                if current_date.weekday() == 4:
                    yield current_date
            else:
                start_end_list = cls.get_trading_intervals()

                for start_end in start_end_list:
                    start = start_end[0]
                    end = start_end[1]

                    current_timestamp = date_and_time(the_date=current_date, the_time=start)
                    end_timestamp = date_and_time(the_date=current_date, the_time=end)

                    while current_timestamp <= end_timestamp:
                        yield current_timestamp
                        current_timestamp = current_timestamp + timedelta(minutes=level.to_minute())

    @classmethod
    def is_open_timestamp(cls, timestamp):
        timestamp = pd.Timestamp(timestamp)
        return is_same_time(timestamp, date_and_time(the_date=timestamp.date(),
                                                     the_time=cls.get_trading_intervals()[0][0]))

    @classmethod
    def is_close_timestamp(cls, timestamp):
        timestamp = pd.Timestamp(timestamp)
        return is_same_time(timestamp, date_and_time(the_date=timestamp.date(),
                                                     the_time=cls.get_trading_intervals()[-1][1]))

    @classmethod
    def is_finished_kdata_timestamp(cls, timestamp: pd.Timestamp, level: IntervalLevel):
        """
        :param timestamp: the timestamp could be recorded in kdata of the level
        :type timestamp: pd.Timestamp
        :param level:
        :type level: zvtm.domain.common.IntervalLevel
        :return:
        :rtype: bool
        """
        timestamp = pd.Timestamp(timestamp)

        for t in cls.get_interval_timestamps(timestamp.date(), timestamp.date(), level=level):
            if is_same_time(t, timestamp):
                return True

        return False

    @classmethod
    def could_short(cls):
        """
        whether could be shorted

        :return:
        """
        return False

    @classmethod
    def get_trading_t(cls):
        """
        0 means t+0
        1 means t+1

        :return:
        """
        return 1


class ActorEntity(Entity):
    pass


class NormalEntityMixin(TradableEntity):
    # the record created time in db
    created_timestamp = Column(DateTime, default=pd.Timestamp.now())
    # the record updated time in db, some recorder would check it for whether need to refresh
    updated_timestamp = Column(DateTime)


class Portfolio(TradableEntity):
    @classmethod
    def get_stocks(cls,
                   code=None, codes=None, ids=None, timestamp=now_pd_timestamp(), provider=None):
        """
        the publishing policy of portfolio positions is different for different types,
        overwrite this function for get the holding stocks in specific date

        :param code: portfolio(etf/block/index...) code
        :param codes: portfolio(etf/block/index...) codes
        :param ids: portfolio(etf/block/index...) ids
        :param timestamp: the date of the holding stocks
        :param provider: the data provider
        :return:
        """
        from zvtm.contract.api import get_schema_by_name
        schema_str = f'{cls.__name__}Stock'
        portfolio_stock = get_schema_by_name(schema_str)
        return portfolio_stock.query_data(provider=provider, code=code, codes=codes, timestamp=timestamp, ids=ids)


# 组合(Fund,Etf,Index,Block等)和个股(Stock)的关系 应该继承自该类
# 该基础类可以这样理解:
# entity为组合本身,其包含了stock这种entity,timestamp为持仓日期,从py的"你知道你在干啥"的哲学出发，不加任何约束
class PortfolioStock(Mixin):
    # portfolio标的类型
    entity_type = Column(String(length=64))
    # portfolio所属交易所
    exchange = Column(String(length=32))
    # portfolio编码
    code = Column(String(length=64))
    # portfolio名字
    name = Column(String(length=128))

    stock_id = Column(String(length=128))
    stock_code = Column(String(length=64))
    stock_name = Column(String(length=128))


# 支持时间变化,报告期标的调整
class PortfolioStockHistory(PortfolioStock):
    # 报告期,season1,half_year,season3,year
    report_period = Column(String(length=32))
    # 3-31,6-30,9-30,12-31
    report_date = Column(DateTime)

    # 占净值比例
    proportion = Column(Float)
    # 持有股票的数量
    shares = Column(Float)
    # 持有股票的市值
    market_cap = Column(Float)


# 交易标的和参与者的关系应该继承自该类, meet,遇见,恰如其分的诠释参与者和交易标的的关系
# 市场就是参与者与交易标的的关系，类的命名规范为{Entity}{relation}{Entity}，entity_id代表"所"为的entity,"受"者entity以具体类别的id命名
# 比如StockTopTenHolder:TradableMeetActor中entity_id和actor_id,分别代表股票和股东
class TradableMeetActor(Mixin):
    # tradable code
    code = Column(String(length=64))
    # tradable name
    name = Column(String(length=128))

    actor_id = Column(String(length=128))
    actor_type = Column(String(length=128))
    actor_code = Column(String(length=64))
    actor_name = Column(String(length=128))


# 也可以"所"为参与者，"受"为标的
class ActorMeetTradable(Mixin):
    # actor code
    code = Column(String(length=64))
    # actor name
    name = Column(String(length=128))

    tradable_id = Column(String(length=128))
    tradable_type = Column(String(length=128))
    tradable_code = Column(String(length=64))
    tradable_name = Column(String(length=128))


# the __all__ is generated
__all__ = ['Mixin', 'NormalMixin', 'Entity', 'TradableEntity', 'ActorEntity', 'NormalEntityMixin', 'Portfolio',
           'PortfolioStock', 'PortfolioStockHistory', 'TradableMeetActor', 'ActorMeetTradable']



# -*- coding: utf-8 -*-
# this file is generated by gen_kdata_schema function, dont't change it
from sqlalchemy.orm import declarative_base

from zvtm.contract.register import register_schema
from zvtm.domain.quotes import IndexKdataCommon

KdataBase = declarative_base()


class Index1wkKdata(KdataBase, IndexKdataCommon):
    __tablename__ = 'index_1wk_kdata'


register_schema(providers=['em', 'sina','joinquant'], db_name='index_1wk_kdata', schema_base=KdataBase, entity_type='index')

# the __all__ is generated
__all__ = ['Index1wkKdata']


# -*- coding: utf-8 -*-
class TraderError(Exception):
    """Base class for exceptions in this module."""
    pass


class InvalidOrderParamError(TraderError):
    def __init__(self, message="invalid order param"):
        self.message = message


class NotEnoughMoneyError(TraderError):
    def __init__(self, message="not enough money"):
        self.message = message


class NotEnoughPositionError(TraderError):
    def __init__(self, message="not enough position"):
        self.message = message


class InvalidOrderError(TraderError):
    def __init__(self, message="invalid order"):
        self.message = message


class WrongKdataError(TraderError):
    def __init__(self, message="wrong kdata"):
        self.message = message
# the __all__ is generated
__all__ = ['TraderError', 'InvalidOrderParamError', 'NotEnoughMoneyError', 'NotEnoughPositionError', 'InvalidOrderError', 'WrongKdataError']


# -*- coding: utf-8 -*-
from enum import Enum

from zvtm.domain import Stock, BlockStock, Block
from zvtm.tag.dataset.stock_tags import StockTags
from zvtm.tag.tag import StockTagger


class CycleTag(Enum):
    # 强周期
    strong_cycle = 'strong_cycle'
    # 弱周期
    weak_cycle = 'weak_cycle'
    # 非周期，受周期影响不大，比如消费，医药
    non_cycle = 'non_cycle'


# 这里用的是东财的行业分类
# 数据来源
# Block.record_data(provider='eastmoney')
# Block.query_data(provider='eastmoney', filters=[Block.category==industry])
# BlockStock.record_data(provider='eastmoney')

# 这不是一个严格的分类，好像也不需要太严格
cycle_map_industry = {
    CycleTag.strong_cycle: ['有色金属', '水泥建材', '化工行业', '输配电气', '化纤行业',
                            '钢铁行业', '煤炭采选', '化肥行业', '贵金属', '船舶制造',
                            '房地产', '石油行业', '港口水运', '材料行业', '工程建设',
                            '包装材料', '券商信托', '机械行业', '金属制品', '塑胶制品',
                            '环保工程', '玻璃陶瓷'],
    CycleTag.weak_cycle: ['电力行业', '民航机场', '家电行业', '旅游酒店', '银行',
                          '保险', '高速公路', '电子信息', '通讯行业', '多元金融',
                          '电子元件', '国际贸易', '珠宝首饰', '交运物流', '航天航空',
                          '交运设备', '汽车行业', '专用设备', '园林工程', '造纸印刷',
                          '安防设备', '装修装饰', '木业家具'],
    CycleTag.non_cycle: ['食品饮料', '酿酒行业', '医疗行业', '医药制造', '文教休闲',
                         '软件服务', '商业百货', '文化传媒', '农牧饲渔', '公用事业',
                         '纺织服装', '综合行业', '仪器仪表', '电信运营', '工艺商品',
                         '农药兽药']
}


def get_cycle_tag(industry_name):
    for cycle_tag in cycle_map_industry:
        if industry_name in cycle_map_industry.get(cycle_tag):
            return cycle_tag.name


class CycleTagger(StockTagger):
    def tag(self, timestamp):
        stock_df = Stock.query_data(filters=[Stock.list_date <= timestamp], index='entity_id')
        block_df = Block.query_data(provider='eastmoney', filters=[Block.category == 'industry'], index='entity_id')
        block_ids = block_df.index.tolist()
        block_stock_df = BlockStock.query_data(provider='eastmoney',
                                               entity_ids=block_ids,
                                               filters=[BlockStock.stock_id.in_(stock_df.index.tolist())],
                                               index='stock_id')
        block_stock_df['cycle_tag'] = block_stock_df['name'].apply(lambda name: get_cycle_tag(name))
        strong_cycle_stocks = block_stock_df[block_stock_df.cycle_tag == 'strong_cycle']['stock_id']
        weak_cycle_stocks = block_stock_df[block_stock_df.cycle_tag == 'weak_cycle']['stock_id']
        non_cycle_stocks = block_stock_df[block_stock_df.cycle_tag == 'non_cycle']['stock_id']

        strong_cycle_domains = self.get_tag_domains(entity_ids=strong_cycle_stocks, timestamp=timestamp,
                                                    cycle_tag=CycleTag.strong_cycle.value)
        weak_cycle_domains = self.get_tag_domains(entity_ids=weak_cycle_stocks, timestamp=timestamp,
                                                  cycle_tag=CycleTag.weak_cycle.value)
        non_cycle_domains = self.get_tag_domains(entity_ids=non_cycle_stocks, timestamp=timestamp,
                                                 cycle_tag=CycleTag.non_cycle.value)

        self.session.add_all(strong_cycle_domains)
        self.session.add_all(weak_cycle_domains)
        self.session.add_all(non_cycle_domains)
        self.session.commit()


if __name__ == '__main__':
    CycleTagger().run()
    print(StockTags.query_data(start_timestamp='2021-08-31', filters=[StockTags.cycle_tag != None]))



# -*- coding: utf-8 -*-
import os
import string

from pkg_resources import resource_string

from zvtm.utils.file_utils import list_all_files


def all_tpls(project: str, entity_type: str):
    """
    return list of templates(location,Template)

    :param project:
    :return:
    """
    tpl_dir = os.path.join(os.path.dirname(__file__))
    tpl_files = list_all_files(tpl_dir, ext='template', return_base_name=True)
    tpls = []
    for tpl in tpl_files:
        data = resource_string(__name__, tpl)
        file_location = os.path.splitext(os.path.basename(tpl))[0]
        # we assure that line endings are converted to '\n' for all OS
        data = data.decode(encoding="utf-8").replace(os.linesep, "\n")

        # change path for specific file
        # domain
        if file_location == 'kdata_common.py':
            file_location = f'{project}/domain/quotes/__init__.py'
        elif file_location == 'meta.py':
            file_location = f'{project}/domain/{entity_type}_meta.py'
        # recorder
        elif file_location == 'kdata_recorder.py':
            file_location = f'{project}/recorders/{entity_type}_kdata_recorder.py'
        elif file_location == 'meta_recorder.py':
            file_location = f'{project}/recorders/{entity_type}_meta_recorder.py'
        # fill script
        elif file_location == 'fill_project.py':
            file_location = f'{project}/fill_project.py'
        # tests
        elif file_location == 'test_pass.py':
            file_location = f'tests/test_pass.py'
        elif file_location == 'pkg_init.py':
            file_location = f'{project}/__init__.py'

        tpls.append((file_location, string.Template(data)))
    return tpls


# the __all__ is generated
__all__ = ['all_tpls']



# -*- coding: utf-8 -*-
from sqlalchemy import Column, Float, Integer

from zvtm.contract import Mixin


class MaStatsFactorCommon(Mixin):
    open = Column(Float)
    close = Column(Float)
    high = Column(Float)
    low = Column(Float)
    turnover = Column(Float)

    ma5 = Column(Float)
    ma10 = Column(Float)

    ma34 = Column(Float)
    ma55 = Column(Float)
    ma89 = Column(Float)
    ma144 = Column(Float)

    ma120 = Column(Float)
    ma250 = Column(Float)

    vol_ma30 = Column(Float)

    live = Column(Integer)
    count = Column(Integer)
    distance = Column(Float)
    area = Column(Float)


# the __all__ is generated
__all__ = ['MaStatsFactorCommon']


# -*- coding: utf-8 -*-
import logging
import time

import pandas as pd
import requests

from zvtm.domain import IndexCategory
from zvtm.recorders.consts import DEFAULT_HEADER
from zvtm.utils import to_pd_timestamp

logger = logging.getLogger(__name__)

original_page_url = 'http://www.cnindex.com.cn/zh_indices/sese/index.html?act_menu=1&index_type=-1'
url = 'http://www.cnindex.net.cn/index/indexList?channelCode={}&rows=1000&pageNum=1'

# 中证指数 抓取 风格指数 行业指数 规模指数 基金指数
cni_category_map_url = {
    IndexCategory.style: url.format('202'),
    IndexCategory.industry: url.format('201'),
    IndexCategory.scope: url.format('200'),
    IndexCategory.fund: url.format('207'),
}

# 深证指数 只取规模指数
sz_category_map_url = {
    IndexCategory.scope: url.format('100'),
}


def _get_resp_data(resp: requests.Response):
    resp.raise_for_status()
    return resp.json()['data']


def get_cn_index(index_type='cni', category=IndexCategory.style):
    if index_type == 'cni':
        category_map_url = cni_category_map_url
    elif index_type == 'sz':
        category_map_url = sz_category_map_url
    else:
        logger.error(f'not support index_type: {index_type}')
        assert False

    requests_session = requests.Session()

    url = category_map_url.get(category)

    resp = requests_session.get(url, headers=DEFAULT_HEADER)

    results = _get_resp_data(resp)['rows']
    # e.g
    # amount: 277743699997.9
    # closeingPoint: 6104.7592
    # docchannel: 1039
    # freeMarketValue: 10794695531696.15
    # id: 142
    # indexcode: "399370"
    # indexename: "CNI Growth"
    # indexfullcname: "国证1000成长指数"
    # indexfullename: "CNI 1000 Growth Index"
    # indexname: "国证成长"
    # indexsource: "1"
    # indextype: "202"
    # pb: 5.34
    # peDynamic: 29.8607
    # peStatic: 33.4933
    # percent: 0.0022
    # prefixmonth: null
    # realtimemarket: "1"
    # remark: ""
    # sampleshowdate: null
    # samplesize: 332
    # showcnindex: "1"
    # totalMarketValue: 23113641352198.32
    the_list = []

    logger.info(f'category: {category} ')
    logger.info(f'results: {results} ')
    for i, result in enumerate(results):
        logger.info(f'to {i}/{len(results)}')
        code = result['indexcode']
        info_resp = requests_session.get(f'http://www.cnindex.net.cn/index-intro?indexcode={code}')
        # fbrq: "2010-01-04"
        # jd: 1000
        # jr: "2002-12-31"
        # jsfs: "自由流通市值"
        # jsjj: "国证成长由国证1000指数样本股中成长风格突出的股票组成，为投资者提供更丰富的指数化投资工具。"
        # qzsx: null
        # typl: 2
        # xyfw: "沪深A股"
        # xygz: "在国证1000指数样本股中，选取主营业务收入增长率、净利润增长率和净资产收益率综合排名前332只"
        index_info = _get_resp_data(info_resp)
        name = result['indexname']
        entity_id = f'index_sz_{code}'
        index_item = {
            'id': entity_id,
            'entity_id': entity_id,
            'timestamp': to_pd_timestamp(index_info['jr']),
            'entity_type': 'index',
            'exchange': 'sz',
            'code': code,
            'name': name,
            'category': category.value,
            'list_date': to_pd_timestamp(index_info['fbrq']),
            'base_point': index_info['jd'],
            'publisher': 'cnindex'
        }
        logger.info(index_item)
        the_list.append(index_item)
        time.sleep(3)
    if the_list:
        return pd.DataFrame.from_records(the_list)


if __name__ == '__main__':
    df = get_cn_index()
    print(df)
# the __all__ is generated
__all__ = ['get_cn_index']


# -*- coding: utf-8 -*-
# the __all__ is generated
__all__ = []

# __init__.py structure:
# common code of the package
# export interface in __all__ which contains __all__ of its sub modules

# import all from submodule eastmoney_dividend_detail_recorder
from .eastmoney_dividend_detail_recorder import *
from .eastmoney_dividend_detail_recorder import __all__ as _eastmoney_dividend_detail_recorder_all
__all__ += _eastmoney_dividend_detail_recorder_all

# import all from submodule eastmoney_dividend_financing_recorder
from .eastmoney_dividend_financing_recorder import *
from .eastmoney_dividend_financing_recorder import __all__ as _eastmoney_dividend_financing_recorder_all
__all__ += _eastmoney_dividend_financing_recorder_all

# import all from submodule eastmoney_rights_issue_detail_recorder
from .eastmoney_rights_issue_detail_recorder import *
from .eastmoney_rights_issue_detail_recorder import __all__ as _eastmoney_rights_issue_detail_recorder_all
__all__ += _eastmoney_rights_issue_detail_recorder_all

# import all from submodule eastmoney_spo_detail_recorder
from .eastmoney_spo_detail_recorder import *
from .eastmoney_spo_detail_recorder import __all__ as _eastmoney_spo_detail_recorder_all
__all__ += _eastmoney_spo_detail_recorder_all


# -*- coding: utf-8 -*-

import logging
import math
from typing import List

from zvtm.api import get_kdata, get_kdata_schema
from zvtm.api.trader_info_api import get_trader_info, clear_trader
from zvtm.contract import IntervalLevel, TradableEntity, AdjustType
from zvtm.contract.api import get_db_session, decode_entity_id
from zvtm.domain.trader_info import AccountStats, Position, Order, TraderInfo
from zvtm.trader import TradingSignalType, TradingListener, TradingSignal
from zvtm.trader.errors import NotEnoughMoneyError, InvalidOrderError, NotEnoughPositionError, InvalidOrderParamError, \
    WrongKdataError
from zvtm.utils.pd_utils import pd_is_not_null
from zvtm.utils.time_utils import to_pd_timestamp, to_time_str, TIME_FORMAT_ISO8601, is_same_date
from zvtm.utils.utils import fill_domain_from_dict

ORDER_TYPE_LONG = 'order_long'
ORDER_TYPE_SHORT = 'order_short'
ORDER_TYPE_CLOSE_LONG = 'order_close_long'
ORDER_TYPE_CLOSE_SHORT = 'order_close_short'

from marshmallow_sqlalchemy import SQLAlchemyAutoSchema


# FIXME:better way for schema<->domain,now just dump to schema and use dict['field'] for operation
class AccountDayStatsSchema(SQLAlchemyAutoSchema):
    class Meta:
        model = AccountStats
        include_relationships = True


class PositionSchema(SQLAlchemyAutoSchema):
    class Meta:
        model = Position
        include_relationships = True


account_stats_schema = AccountDayStatsSchema()
position_schema = PositionSchema()


class AccountService(TradingListener):
    logger = logging.getLogger(__name__)

    def get_current_position(self, entity_id):
        """
        overwrite it to provide your real position

        :param entity_id:
        """
        pass

    def get_current_account(self):
        pass

    def order(self, entity_id, current_price, current_timestamp, order_amount=0, order_pct=1.0, order_price=0,
              order_type=ORDER_TYPE_LONG, order_money=0):
        pass

    # 开多,对于某些品种只能开多，比如中国股票
    def buy(self, entity_id, current_price, current_timestamp, order_amount=0, order_pct=1.0, order_price=0,
            order_money=0):
        self.order(entity_id, current_price, current_timestamp, order_amount, order_pct, order_price,
                   order_type=ORDER_TYPE_LONG, order_money=order_money)

    # 开空
    def sell(self, entity_id, current_price, current_timestamp, order_amount=0, order_pct=1.0, order_price=0,
             order_money=0):
        self.order(entity_id, current_price, current_timestamp, order_amount, order_pct, order_price,
                   order_type=ORDER_TYPE_SHORT, order_money=order_money)

    # 平多
    def close_long(self, entity_id, current_price, current_timestamp, order_amount=0, order_pct=1.0, order_price=0,
                   order_money=0):
        self.order(entity_id, current_price, current_timestamp, order_amount, order_pct, order_price,
                   order_type=ORDER_TYPE_CLOSE_LONG, order_money=order_money)

    # 平空
    def close_short(self, entity_id, current_price, current_timestamp, order_amount=0, order_pct=1.0, order_price=0,
                    order_money=0):
        self.order(entity_id, current_price, current_timestamp, order_amount, order_pct, order_price,
                   order_type=ORDER_TYPE_CLOSE_SHORT, order_money=order_money)

    @staticmethod
    def trading_signal_to_order_type(trading_signal_type):
        if trading_signal_type == TradingSignalType.open_long:
            return ORDER_TYPE_LONG
        if trading_signal_type == TradingSignalType.open_short:
            return ORDER_TYPE_SHORT
        if trading_signal_type == TradingSignalType.close_long:
            return ORDER_TYPE_CLOSE_LONG
        if trading_signal_type == TradingSignalType.close_short:
            return ORDER_TYPE_CLOSE_SHORT


class SimAccountService(AccountService):

    def __init__(self,
                 entity_schema: TradableEntity,
                 trader_name,
                 timestamp,
                 provider=None,
                 level=IntervalLevel.LEVEL_1DAY,
                 base_capital=1000000,
                 buy_cost=0.001,
                 sell_cost=0.001,
                 slippage=0.001,
                 rich_mode=True,
                 adjust_type: AdjustType = None,
                 keep_history=False,
                 real_time=False,
                 kdata_use_begin_time=False):
        self.entity_schema = entity_schema
        self.base_capital = base_capital
        self.buy_cost = buy_cost
        self.sell_cost = sell_cost
        self.slippage = slippage
        self.rich_mode = rich_mode
        self.adjust_type = adjust_type
        self.trader_name = trader_name

        self.session = get_db_session('zvt', data_schema=TraderInfo)
        self.provider = provider
        self.level = level
        self.start_timestamp = timestamp
        self.keep_history = keep_history
        self.real_time = real_time
        self.kdata_use_begin_time = kdata_use_begin_time

        self.account = None
        self.account = self.init_account()

        account_info = f'init_account,holding size:{len(self.account.positions)} profit:{self.account.profit} input_money:{self.account.input_money} ' \
                       f'cash:{self.account.cash} value:{self.account.value} all_value:{self.account.all_value}'
        self.logger.info(account_info)

    def input_money(self, money=1000000):
        self.account.input_money += money
        self.account.cash += money

    def clear_account(self):
        trader_info = get_trader_info(session=self.session, trader_name=self.trader_name, return_type='domain',
                                      limit=1)

        if trader_info:
            self.logger.warning("trader:{} has run before,old result would be deleted".format(self.trader_name))
            clear_trader(session=self.session, trader_name=self.trader_name)

    def init_account(self) -> AccountStats:
        # 清除历史数据
        if not self.keep_history:
            self.clear_account()

        # 读取之前保存的账户
        if self.keep_history:
            self.account = self.load_account()
            if self.account:
                return self.account

        # init trader info
        entity_type = self.entity_schema.__name__.lower()
        sim_account = TraderInfo(id=self.trader_name,
                                 entity_id=f'trader_zvt_{self.trader_name}',
                                 timestamp=self.start_timestamp,
                                 trader_name=self.trader_name,
                                 entity_type=entity_type,
                                 start_timestamp=self.start_timestamp,
                                 provider=self.provider,
                                 level=self.level.value,
                                 real_time=self.real_time,
                                 kdata_use_begin_time=self.kdata_use_begin_time,
                                 kdata_adjust_type=self.adjust_type.value)
        self.session.add(sim_account)
        self.session.commit()

        return AccountStats(entity_id=f'trader_zvt_{self.trader_name}',
                            timestamp=self.start_timestamp,
                            trader_name=self.trader_name,
                            cash=self.base_capital,
                            input_money=self.base_capital,
                            all_value=self.base_capital,
                            value=0,
                            closing=False)

    def load_account(self) -> AccountStats:
        records = AccountStats.query_data(filters=[AccountStats.trader_name == self.trader_name],
                                          order=AccountStats.timestamp.desc(), limit=1, return_type='domain')
        if not records:
            return self.account
        latest_record: AccountStats = records[0]

        # create new orm object from latest record
        account_dict = account_stats_schema.dump(latest_record)
        del account_dict['id']
        del account_dict['positions']
        account = AccountStats()
        fill_domain_from_dict(account, account_dict)

        positions: List[Position] = []
        for position_domain in latest_record.positions:
            position_dict = position_schema.dump(position_domain)
            self.logger.debug('current position:{}'.format(position_dict))
            del position_dict['id']
            del position_dict['account_stats']
            position = Position()
            fill_domain_from_dict(position, position_dict)
            positions.append(position)

        account.positions = positions

        return account

    def on_trading_open(self, timestamp):
        self.logger.info('on_trading_open:{}'.format(timestamp))
        if is_same_date(timestamp, self.start_timestamp):
            return
        self.account = self.load_account()

    def on_trading_error(self, timestamp, error):
        pass

    def on_trading_finish(self, timestamp):
        pass

    def on_trading_signals(self, trading_signals: List[TradingSignal]):
        for trading_signal in trading_signals:
            try:
                self.on_trading_signal(trading_signal)
            except Exception as e:
                self.logger.exception(e)
                self.on_trading_error(timestamp=trading_signal.happen_timestamp, error=e)

    def on_trading_signal(self, trading_signal: TradingSignal):
        entity_id = trading_signal.entity_id
        happen_timestamp = trading_signal.happen_timestamp
        order_type = AccountService.trading_signal_to_order_type(trading_signal.trading_signal_type)
        trading_level = trading_signal.trading_level.value
        if order_type:
            try:
                kdata = get_kdata(provider=self.provider, entity_id=entity_id, level=trading_level,
                                  start_timestamp=happen_timestamp, end_timestamp=happen_timestamp,
                                  limit=1, adjust_type=self.adjust_type)
            except Exception as e:
                self.logger.error(e)
                raise WrongKdataError("could not get kdata")

            if pd_is_not_null(kdata):
                entity_type, _, _ = decode_entity_id(kdata['entity_id'][0])

                the_price = kdata['close'][0]

                if the_price:
                    self.order(entity_id=entity_id, current_price=the_price,
                               current_timestamp=happen_timestamp, order_pct=trading_signal.position_pct,
                               order_money=trading_signal.order_money,
                               order_type=order_type)
                else:
                    self.logger.warning(
                        'ignore trading signal,wrong kdata,entity_id:{},timestamp:{},kdata:{}'.format(entity_id,
                                                                                                      happen_timestamp,
                                                                                                      kdata.to_dict(
                                                                                                          orient='records')))

            else:
                self.logger.warning(
                    'ignore trading signal,could not get kdata,entity_id:{},timestamp:{}'.format(entity_id,
                                                                                                 happen_timestamp))

    def on_trading_close(self, timestamp):
        self.logger.info('on_trading_close:{}'.format(timestamp))
        # remove the empty position
        self.account.positions = [position for position in self.account.positions if
                                  position.long_amount > 0 or position.short_amount > 0]

        # clear the data which need recomputing
        the_id = '{}_{}'.format(self.trader_name, to_time_str(timestamp, TIME_FORMAT_ISO8601))

        self.account.value = 0
        self.account.all_value = 0
        for position in self.account.positions:
            entity_type, _, _ = decode_entity_id(position.entity_id)
            data_schema = get_kdata_schema(entity_type, level=IntervalLevel.LEVEL_1DAY, adjust_type=self.adjust_type)

            kdata = get_kdata(provider=self.provider, level=IntervalLevel.LEVEL_1DAY, entity_id=position.entity_id,
                              order=data_schema.timestamp.desc(),
                              end_timestamp=timestamp, limit=1, adjust_type=self.adjust_type)

            closing_price = kdata['close'][0]

            position.available_long = position.long_amount
            position.available_short = position.short_amount

            if closing_price:
                if (position.long_amount is not None) and position.long_amount > 0:
                    position.value = position.long_amount * closing_price
                    self.account.value += position.value
                elif (position.short_amount is not None) and position.short_amount > 0:
                    position.value = 2 * (position.short_amount * position.average_short_price)
                    position.value -= position.short_amount * closing_price
                    self.account.value += position.value

                # refresh profit
                position.profit = (closing_price - position.average_long_price) \
                                  * position.long_amount
                position.profit_rate = position.profit / (
                        position.average_long_price * position.long_amount)

            else:
                self.logger.warning(
                    'could not refresh close value for position:{},timestamp:{}'.format(position.entity_id,
                                                                                        timestamp))

            position.id = '{}_{}_{}'.format(self.trader_name, position.entity_id,
                                            to_time_str(timestamp, TIME_FORMAT_ISO8601))
            position.timestamp = to_pd_timestamp(timestamp)
            position.account_stats_id = the_id

        self.account.id = the_id
        self.account.all_value = self.account.value + self.account.cash
        self.account.closing = True
        self.account.timestamp = to_pd_timestamp(timestamp)
        self.account.profit = (self.account.all_value - self.account.input_money) / self.account.input_money

        self.session.add(self.account)
        self.session.commit()
        account_info = f'on_trading_close,holding size:{len(self.account.positions)} profit:{self.account.profit} input_money:{self.account.input_money} ' \
                       f'cash:{self.account.cash} value:{self.account.value} all_value:{self.account.all_value}'
        self.logger.info(account_info)

    def get_current_position(self, entity_id) -> Position:
        """
        get current position to decide whether order could make

        :param entity_id:
        :type entity_id: str
        :return:
        :rtype: None
        """
        for position in self.account.positions:
            if position.entity_id == entity_id:
                return position
        return None

    def get_current_account(self):
        return self.account

    def update_position(self, current_position, order_amount, current_price, order_type, timestamp):
        """

        :param timestamp:
        :type timestamp:
        :param current_position:
        :type current_position: Position
        :param order_amount:
        :type order_amount:
        :param current_price:
        :type current_price:
        :param order_type:
        :type order_type:
        """
        if order_type == ORDER_TYPE_LONG:
            need_money = (order_amount * current_price) * (1 + self.slippage + self.buy_cost)
            if self.account.cash < need_money:
                if self.rich_mode:
                    self.input_money()
                else:
                    raise NotEnoughMoneyError()

            self.account.cash -= need_money

            # 计算平均价
            long_amount = current_position.long_amount + order_amount
            if long_amount == 0:
                current_position.average_long_price = 0
            current_position.average_long_price = (current_position.average_long_price * current_position.long_amount
                                                   + current_price * order_amount) / long_amount

            current_position.long_amount = long_amount

            if current_position.trading_t == 0:
                current_position.available_long += order_amount

        elif order_type == ORDER_TYPE_SHORT:
            need_money = (order_amount * current_price) * (1 + self.slippage + self.buy_cost)
            if self.account.cash < need_money:
                if self.rich_mode:
                    self.input_money()
                else:
                    raise NotEnoughMoneyError()

            self.account.cash -= need_money

            short_amount = current_position.short_amount + order_amount
            current_position.average_short_price = (current_position.average_short_price * current_position.short_amount
                                                    + current_price * order_amount) / short_amount

            current_position.short_amount = short_amount

            if current_position.trading_t == 0:
                current_position.available_short += order_amount

        elif order_type == ORDER_TYPE_CLOSE_LONG:
            self.account.cash += (order_amount * current_price * (1 - self.slippage - self.sell_cost))
            # FIXME:如果没卖完，重新计算计算平均价

            current_position.available_long -= order_amount
            current_position.long_amount -= order_amount

        elif order_type == ORDER_TYPE_CLOSE_SHORT:
            self.account.cash += 2 * (order_amount * current_position.average_short_price)
            self.account.cash -= order_amount * current_price * (1 + self.slippage + self.sell_cost)

            current_position.available_short -= order_amount
            current_position.short_amount -= order_amount

        # save the order info to db
        order_id = '{}_{}_{}_{}'.format(self.trader_name, order_type, current_position.entity_id,
                                        to_time_str(timestamp, TIME_FORMAT_ISO8601))
        order = Order(id=order_id,
                      timestamp=to_pd_timestamp(timestamp),
                      trader_name=self.trader_name,
                      entity_id=current_position.entity_id,
                      order_price=current_price,
                      order_amount=order_amount,
                      order_type=order_type,
                      level=self.level.value,
                      status='success')
        self.session.add(order)
        self.session.commit()

    def order(self, entity_id, current_price, current_timestamp, order_amount=0, order_pct=1.0, order_price=0,
              order_type=ORDER_TYPE_LONG, order_money=0):
        """
        下单

        Parameters
        ----------
        entity_id : str
            交易标的id

        current_price : float
            当前价格

        current_timestamp: timestamp
            下单的时间

        order_amount : int
            数量

        order_pct : float
            使用可用现金(仓位)的百分比,0.0-1.0

        order_price : float
            用于限价交易

        order_type : {ORDER_TYPE_LONG,ORDER_TYPE_SHORT,ORDER_TYPE_CLOSE_LONG,ORDER_TYPE_CLOSE_SHORT}
            交易类型

        Returns

        """

        # 市价交易,就是买卖是"当时"并"一定"能成交的
        # 简单起见，目前只支持这种方式
        if order_price == 0:
            current_position = self.get_current_position(entity_id=entity_id)

            if not current_position:
                trading_t = self.entity_schema.get_trading_t()
                current_position = Position(trader_name=self.trader_name,
                                            entity_id=entity_id,
                                            long_amount=0,
                                            available_long=0,
                                            average_long_price=0,
                                            short_amount=0,
                                            available_short=0,
                                            average_short_price=0,
                                            profit=0,
                                            value=0,
                                            trading_t=trading_t
                                            )
                # add it to latest account
                self.account.positions.append(current_position)

            # 按钱交易
            if order_money > 0:
                # 开多
                if order_type == ORDER_TYPE_LONG:
                    if current_position.short_amount > 0:
                        raise InvalidOrderError("close the short position before open long")

                    if order_money > self.account.cash:
                        if self.rich_mode:
                            self.input_money()
                        else:
                            raise NotEnoughMoneyError()

                    cost = current_price * (1 + self.slippage + self.buy_cost)
                    # 买的数量
                    order_amount = order_money // cost

                    if order_amount < 1:
                        self.logger.error(
                            f'invalid order_money:{order_money}, cost:{cost}, order_amount:{order_amount}')
                        return

                    self.update_position(current_position, order_amount, current_price, order_type,
                                         current_timestamp)
                # 开空
                elif order_type == ORDER_TYPE_SHORT:
                    if current_position.long_amount > 0:
                        raise InvalidOrderError("close the long position before open short")

                    if order_money > self.account.cash:
                        if self.rich_mode:
                            self.input_money()
                        else:
                            raise NotEnoughMoneyError()

                    cost = current_price * (1 + self.slippage + self.buy_cost)

                    order_amount = order_money // cost

                    if order_amount < 1:
                        self.logger.error(
                            f'invalid order_money:{order_money}, cost:{cost}, order_amount:{order_amount}')
                        return
                    self.update_position(current_position, order_amount, current_price, order_type,
                                         current_timestamp)
                else:
                    raise InvalidOrderParamError('close long/short not support order_money')

            # 按数量交易
            elif order_amount > 0:
                # 开多
                if order_type == ORDER_TYPE_LONG:
                    if current_position.short_amount > 0:
                        raise InvalidOrderError("close the short position before open long")

                    self.update_position(current_position, order_amount, current_price, order_type, current_timestamp)
                # 开空
                elif order_type == ORDER_TYPE_SHORT:
                    if current_position.long_amount > 0:
                        raise InvalidOrderError("close the long position before open short")

                    self.update_position(current_position, order_amount, current_price, order_type, current_timestamp)
                # 平多
                elif order_type == ORDER_TYPE_CLOSE_LONG:
                    if current_position.available_long >= order_amount:
                        self.update_position(current_position, order_amount, current_price, order_type,
                                             current_timestamp)
                    else:
                        raise NotEnoughPositionError()
                # 平空
                elif order_type == ORDER_TYPE_CLOSE_SHORT:
                    if current_position.available_short >= order_amount:
                        self.update_position(current_position, order_amount, current_price, order_type,
                                             current_timestamp)
                    else:
                        raise Exception("not enough position")

            # 按仓位比例交易
            elif 0 < order_pct <= 1:
                # 开多
                if order_type == ORDER_TYPE_LONG:
                    if current_position.short_amount > 0:
                        raise InvalidOrderError("close the short position before open long")

                    cost = current_price * (1 + self.slippage + self.buy_cost)
                    want_pay = self.account.cash * order_pct
                    # 买的数量
                    order_amount = want_pay // cost

                    if order_amount < 1:
                        if self.rich_mode:
                            self.input_money()
                            order_amount = max((self.account.cash * order_pct) // cost, 1)
                        else:
                            raise NotEnoughMoneyError()
                    self.update_position(current_position, order_amount, current_price, order_type,
                                         current_timestamp)
                # 开空
                elif order_type == ORDER_TYPE_SHORT:
                    if current_position.long_amount > 0:
                        raise InvalidOrderError("close the long position before open short")

                    cost = current_price * (1 + self.slippage + self.buy_cost)
                    want_pay = self.account.cash * order_pct

                    order_amount = want_pay // cost

                    if order_amount < 1:
                        if self.rich_mode:
                            self.input_money()
                            order_amount = max((self.account.cash * order_pct) // cost, 1)
                        else:
                            raise NotEnoughMoneyError()

                    self.update_position(current_position, order_amount, current_price, order_type,
                                         current_timestamp)

                # 平多
                elif order_type == ORDER_TYPE_CLOSE_LONG:
                    if current_position.available_long > 0:
                        if order_pct == 1.0:
                            order_amount = current_position.available_long
                        else:
                            order_amount = math.floor(current_position.available_long * order_pct)

                        if order_amount != 0:
                            self.update_position(current_position, order_amount, current_price, order_type,
                                                 current_timestamp)
                        else:
                            self.logger.warning(
                                f'{entity_id} available_long:{current_position.available_long} order_pct:{order_pct} order_amount:{order_amount}')
                    else:
                        raise NotEnoughPositionError()
                # 平空
                elif order_type == ORDER_TYPE_CLOSE_SHORT:
                    if current_position.available_short > 0:
                        if order_pct == 1.0:
                            order_amount = current_position.available_short
                        else:
                            order_amount = math.floor(current_position.available_short * order_pct)

                        if order_amount != 0:
                            self.update_position(current_position, order_amount, current_price, order_type,
                                                 current_timestamp)
                        else:
                            self.logger.warning(
                                f'{entity_id} available_long:{current_position.available_long} order_pct:{order_pct} order_amount:{order_amount}')
                    else:
                        raise Exception("not enough position")


# the __all__ is generated
__all__ = ['AccountDayStatsSchema', 'PositionSchema', 'AccountService', 'SimAccountService']



# -*- coding: utf-8 -*-
import logging
import time
import uuid
from typing import List

import pandas as pd
from sqlalchemy.orm import Session

from zvtm.contract import IntervalLevel, Mixin, TradableEntity
from zvtm.contract.api import get_db_session, get_schema_columns
from zvtm.contract.api import get_entities, get_data
from zvtm.contract.base import StatefulService
from zvtm.contract.zvt_info import RecorderState
from zvtm.utils import pd_is_not_null
from zvtm.utils.time_utils import to_pd_timestamp, TIME_FORMAT_DAY, to_time_str, \
    evaluate_size_from_timestamp, is_in_same_interval, now_pd_timestamp, now_time_str
from zvtm.utils.utils import fill_domain_from_dict


class Meta(type):
    def __new__(meta, name, bases, class_dict):
        cls = type.__new__(meta, name, bases, class_dict)
        # register the recorder class to the data_schema
        if hasattr(cls, 'data_schema') and hasattr(cls, 'provider'):
            if cls.data_schema and issubclass(cls.data_schema, Mixin):
                print(f'{cls.__name__}:{cls.data_schema.__name__}')
                cls.data_schema.register_recorder_cls(cls.provider, cls)
        return cls


class Recorder(StatefulService, metaclass=Meta):
    # overwrite them to setup the data you want to record
    provider: str = None
    data_schema: Mixin = None

    # original page url
    original_page_url = None
    # request url
    url = None

    state_schema = RecorderState

    def __init__(self, force_update: bool = False, sleeping_time: int = 10) -> None:
        super().__init__()
        self.logger = logging.getLogger(self.__class__.__name__)

        assert self.provider is not None
        assert self.data_schema is not None
        if self.provider not in self.data_schema.providers:
            self.logger.error(
                f'provider: {self.provider} is not registered for {self.data_schema}({self.data_schema.providers})')
            assert False

        self.force_update = force_update
        self.sleeping_time = sleeping_time

        # using to do db operations
        self.session = get_db_session(provider=self.provider,
                                      data_schema=self.data_schema)

    def run(self):
        raise NotImplementedError

    def sleep(self, seconds=None):
        if seconds:
            sleeping_time = seconds
        else:
            sleeping_time = self.sleeping_time

        if sleeping_time and sleeping_time > 0:
            self.logger.info(f'sleeping {sleeping_time} seconds')
            time.sleep(self.sleeping_time)


class EntityEventRecorder(Recorder):
    # overwrite them to fetch the entity list
    entity_provider: str = None
    entity_schema: TradableEntity = None

    def __init__(self,
                 force_update=False,
                 sleeping_time=10,
                 exchanges=None,
                 entity_ids=None,
                 code=None,
                 codes=None,
                 day_data=False,
                 entity_filters=None,
                 ignore_failed=True) -> None:
        """
        :param code:
        :param ignore_failed:
        :param entity_filters:
        :param exchanges:
        :param entity_ids: set entity_ids or (entity_type,exchanges,codes)
        :param codes:
        :param day_data: one record per day,set to True if you want skip recording it when data of today exist
        :param force_update:
        :param sleeping_time:
        """
        super().__init__(force_update=force_update, sleeping_time=sleeping_time)

        assert self.entity_provider is not None
        assert self.entity_schema is not None

        # setup the entities you want to record
        self.exchanges = exchanges
        if codes is None and code is not None:
            self.codes = [code]
        else:
            self.codes = codes
        self.day_data = day_data

        # set entity_ids or (entity_type,exchanges,codes)
        self.entity_ids = entity_ids
        self.entity_filters = entity_filters
        self.ignore_failed = ignore_failed

        self.entity_session: Session = None
        self.entities: List = None
        self.init_entities()

    def init_entities(self):
        """
        init the entities which we would record data for
        """
        if self.entity_provider == self.provider and self.entity_schema == self.data_schema:
            self.entity_session = self.session
        else:
            self.entity_session = get_db_session(provider=self.entity_provider, data_schema=self.entity_schema)

        if self.day_data:
            df = self.data_schema.query_data(start_timestamp=now_time_str(), columns=['entity_id', 'timestamp'],
                                             provider=self.provider)
            if pd_is_not_null(df):
                entity_ids = df['entity_id'].tolist()
                self.logger.info(f'ignore entity_ids:{entity_ids}')
                if self.entity_filters:
                    self.entity_filters.append(self.entity_schema.entity_id.notin_(entity_ids))
                else:
                    self.entity_filters = [self.entity_schema.entity_id.notin_(entity_ids)]

        # init the entity list
        self.entities = get_entities(session=self.entity_session,
                                     entity_schema=self.entity_schema,
                                     exchanges=self.exchanges,
                                     entity_ids=self.entity_ids,
                                     codes=self.codes,
                                     return_type='domain',
                                     provider=self.entity_provider,
                                     filters=self.entity_filters)


class TimeSeriesDataRecorder(EntityEventRecorder):
    default_size = 2000

    def __init__(self,
                 force_update=False,
                 sleeping_time=5,
                 exchanges=None,
                 entity_ids=None,
                 code=None,
                 codes=None,
                 day_data=False,
                 entity_filters=None,
                 ignore_failed=True,
                 real_time=False,
                 fix_duplicate_way='add',
                 start_timestamp=None,
                 end_timestamp=None) -> None:
        super().__init__(force_update, sleeping_time, exchanges, entity_ids, code=code, codes=codes, day_data=day_data,
                         entity_filters=entity_filters, ignore_failed=ignore_failed)

        self.real_time = real_time
        self.close_hour, self.close_minute = self.entity_schema.get_close_hour_and_minute()
        self.fix_duplicate_way = fix_duplicate_way
        self.start_timestamp = to_pd_timestamp(start_timestamp)
        self.end_timestamp = to_pd_timestamp(end_timestamp)

    def get_latest_saved_record(self, entity):
        order = eval('self.data_schema.{}.desc()'.format(self.get_evaluated_time_field()))

        records = get_data(entity_id=entity.id,
                           provider=self.provider,
                           data_schema=self.data_schema,
                           order=order,
                           limit=1,
                           return_type='domain',
                           session=self.session)
        if records:
            return records[0]
        return None

    def evaluate_start_end_size_timestamps(self, entity):
        # not to list date yet
        if entity.timestamp and (entity.timestamp >= now_pd_timestamp()):
            self.logger.info("ignore entity: {} list date: {}", entity.id, entity.timestamp)
            return entity.timestamp, None, 0, None

        latest_saved_record = self.get_latest_saved_record(entity=entity)

        if latest_saved_record:
            latest_timestamp = eval('latest_saved_record.{}'.format(self.get_evaluated_time_field()))
        else:
            latest_timestamp = entity.timestamp

        if not latest_timestamp:
            return self.start_timestamp, self.end_timestamp, self.default_size, None

        if self.start_timestamp:
            latest_timestamp = max(latest_timestamp, self.start_timestamp)

        size = self.default_size
        if self.end_timestamp:
            if latest_timestamp > self.end_timestamp:
                size = 0

        return latest_timestamp, self.end_timestamp, size, None

    def get_data_map(self):
        """
        {'original_field':('domain_field',transform_func)}
        """
        return {}

    def record(self, entity, start, end, size, timestamps):
        """
        implement the recording logic in this method, should return json or domain list
        :param entity:
        :type entity:
        :param start:
        :type start:
        :param end:
        :type end:
        :param size:
        :type size:
        :param timestamps:
        :type timestamps:
        """
        raise NotImplementedError

    def get_evaluated_time_field(self):
        """
        the timestamp field for evaluating time range of recorder,used in get_latest_saved_record
        """
        return 'timestamp'

    def get_original_time_field(self):
        return 'timestamp'

    def generate_domain_id(self, entity, original_data, time_fmt=TIME_FORMAT_DAY):
        """
        generate domain id from the entity and original data,the default id meaning:entity + event happen time
        :param entity:
        :type entity:
        :param original_data:
        :type original_data:
        :param time_fmt:
        :type time_fmt:
        :return:
        :rtype:
        """
        timestamp = to_time_str(original_data[self.get_original_time_field()], fmt=time_fmt)
        return "{}_{}".format(entity.id, timestamp)

    def generate_domain(self, entity, original_data):
        """
        generate the data_schema instance using entity and original_data,the original_data is from record result
        :param entity:
        :param original_data:
        """

        got_new_data = False

        # if the domain is directly generated in record method, we just return it
        if isinstance(original_data, self.data_schema):
            got_new_data = True
            return got_new_data, original_data

        the_id = self.generate_domain_id(entity, original_data)

        # optional way
        # item = self.session.query(self.data_schema).get(the_id)

        items = get_data(data_schema=self.data_schema, session=self.session, provider=self.provider,
                         entity_id=entity.id, filters=[self.data_schema.id == the_id], return_type='domain')

        if items and not self.force_update:
            self.logger.info('ignore the data {}:{} saved before'.format(self.data_schema, the_id))
            return got_new_data, None

        if not items:
            timestamp_str = original_data[self.get_original_time_field()]
            timestamp = None
            try:
                timestamp = to_pd_timestamp(timestamp_str)
            except Exception as e:
                self.logger.exception(e)

            if 'name' in get_schema_columns(self.data_schema):
                domain_item = self.data_schema(id=the_id,
                                               code=entity.code,
                                               name=entity.name,
                                               entity_id=entity.id,
                                               timestamp=timestamp)
            else:
                domain_item = self.data_schema(id=the_id,
                                               code=entity.code,
                                               entity_id=entity.id,
                                               timestamp=timestamp)
            got_new_data = True
        else:
            domain_item = items[0]

        fill_domain_from_dict(domain_item, original_data, self.get_data_map())
        return got_new_data, domain_item

    def persist(self, entity, domain_list):
        """
        persist the domain list to db
        :param entity:
        :param domain_list:
        """
        if domain_list:
            try:
                if domain_list[0].timestamp >= domain_list[-1].timestamp:
                    first_timestamp = domain_list[-1].timestamp
                    last_timestamp = domain_list[0].timestamp
                else:
                    first_timestamp = domain_list[0].timestamp
                    last_timestamp = domain_list[-1].timestamp
            except:
                first_timestamp = domain_list[0].timestamp
                last_timestamp = domain_list[-1].timestamp

            self.logger.info(
                "persist {} for entity_id:{},time interval:[{},{}]".format(
                    self.data_schema, entity.id, first_timestamp, last_timestamp))

            self.session.add_all(domain_list)
            self.session.commit()

    def on_finish(self):
        try:
            if self.session:
                self.session.close()
                #self.session.dispose()


            if self.entity_session:
                self.entity_session.close()
                #self.entity_session.dospose()
        except Exception as e:
            self.logger.error(e)

    def on_finish_entity(self, entity):
        pass

    def run(self):
        finished_items = []
        unfinished_items = self.entities
        raising_exception = None
        while True:
            count = len(unfinished_items)
            for index, entity_item in enumerate(unfinished_items):
                try:
                    self.logger.info(f'run to {index + 1}/{count}')

                    start_timestamp, end_timestamp, size, timestamps = self.evaluate_start_end_size_timestamps(
                        entity_item)
                    size = int(size)

                    if timestamps:
                        self.logger.info('entity_id:{},evaluate_start_end_size_timestamps result:{},{},{},{}-{}'.format(
                            entity_item.id,
                            start_timestamp,
                            end_timestamp,
                            size,
                            timestamps[0],
                            timestamps[-1]))
                    else:
                        self.logger.info('entity_id:{},evaluate_start_end_size_timestamps result:{},{},{},{}'.format(
                            entity_item.id,
                            start_timestamp,
                            end_timestamp,
                            size,
                            timestamps))

                    # no more to record
                    if size == 0:
                        finished_items.append(entity_item)
                        self.logger.info(
                            "finish recording {} for entity_id:{},latest_timestamp:{}".format(
                                self.data_schema,
                                entity_item.id,
                                start_timestamp))
                        self.on_finish_entity(entity_item)
                        continue

                    # sleep for a while to next entity
                    if index != 0:
                        self.sleep()

                    original_list = self.record(entity_item, start=start_timestamp, end=end_timestamp, size=size,
                                                timestamps=timestamps)

                    all_duplicated = True

                    if original_list:
                        domain_list = []
                        for original_item in original_list:
                            got_new_data, domain_item = self.generate_domain(entity_item, original_item)

                            if got_new_data:
                                all_duplicated = False

                            # handle the case  generate_domain_id generate duplicate id
                            if domain_item:
                                duplicate = [item for item in domain_list if item.id == domain_item.id]
                                if duplicate:
                                    # regenerate the id
                                    if self.fix_duplicate_way == 'add':
                                        domain_item.id = "{}_{}".format(domain_item.id, uuid.uuid1())
                                    # ignore
                                    else:
                                        self.logger.info(f'ignore original duplicate item:{domain_item.id}')
                                        continue

                                domain_list.append(domain_item)

                        if domain_list:
                            self.persist(entity_item, domain_list)
                        else:
                            self.logger.info('just got {} duplicated data in this cycle'.format(len(original_list)))

                    # could not get more data
                    entity_finished = False
                    if not original_list or all_duplicated:
                        # not realtime
                        if not self.real_time:
                            entity_finished = True

                        # realtime and to the close time
                        if self.real_time and \
                                (self.close_hour is not None) and \
                                (self.close_minute is not None):
                            current_timestamp = pd.Timestamp.now()
                            if current_timestamp.hour >= self.close_hour:
                                if current_timestamp.minute - self.close_minute >= 5:
                                    self.logger.info(
                                        '{} now is the close time:{}'.format(entity_item.id, current_timestamp))

                                    entity_finished = True

                    # add finished entity to finished_items
                    if entity_finished:
                        finished_items.append(entity_item)

                        latest_saved_record = self.get_latest_saved_record(entity=entity_item)
                        if latest_saved_record:
                            start_timestamp = eval('latest_saved_record.{}'.format(self.get_evaluated_time_field()))

                        self.logger.info(
                            "finish recording {} for entity_id:{},latest_timestamp:{}".format(
                                self.data_schema,
                                entity_item.id,
                                start_timestamp))
                        self.on_finish_entity(entity_item)
                        continue

                except Exception as e:
                    self.logger.exception(
                        "recording data for entity_id:{},{},error:{}".format(entity_item.id, self.data_schema, e))
                    raising_exception = e
                    finished_items = unfinished_items
                    break

            unfinished_items = set(unfinished_items) - set(finished_items)

            if len(unfinished_items) == 0:
                break

        self.on_finish()

        if raising_exception:
            raise raising_exception


class FixedCycleDataRecorder(TimeSeriesDataRecorder):
    def __init__(self,
                 force_update=True,
                 sleeping_time=10,
                 exchanges=None,
                 entity_ids=None,
                 code=None,
                 codes=None,
                 day_data=False,
                 entity_filters=None,
                 ignore_failed=True,
                 real_time=False,
                 fix_duplicate_way='ignore',
                 start_timestamp=None,
                 end_timestamp=None,
                 level=IntervalLevel.LEVEL_1DAY,
                 kdata_use_begin_time=False,
                 one_day_trading_minutes=24 * 60) -> None:
        super().__init__(force_update, sleeping_time, exchanges, entity_ids, code=code, codes=codes, day_data=day_data,
                         entity_filters=entity_filters, ignore_failed=ignore_failed, real_time=real_time,
                         fix_duplicate_way=fix_duplicate_way, start_timestamp=start_timestamp,
                         end_timestamp=end_timestamp)

        self.level = IntervalLevel(level)
        self.kdata_use_begin_time = kdata_use_begin_time
        self.one_day_trading_minutes = one_day_trading_minutes

    def get_latest_saved_record(self, entity):
        order = eval('self.data_schema.{}.desc()'.format(self.get_evaluated_time_field()))

        # 对于k线这种数据，最后一个记录有可能是没完成的，所以取两个
        # 同一周期内只保留最新的一个数据
        records = get_data(entity_id=entity.id,
                           provider=self.provider,
                           data_schema=self.data_schema,
                           order=order,
                           limit=2,
                           return_type='domain',
                           session=self.session,
                           level=self.level)
        if records:
            # delete unfinished kdata
            if len(records) == 2:
                if is_in_same_interval(t1=records[0].timestamp, t2=records[1].timestamp, level=self.level):
                    self.session.delete(records[1])
                    self.session.flush()
            return records[0]
        return None

    def evaluate_start_end_size_timestamps(self, entity):
        # not to list date yet
        if entity.timestamp and (entity.timestamp >= now_pd_timestamp()):
            return entity.timestamp, None, 0, None

        # get latest record
        latest_saved_record = self.get_latest_saved_record(entity=entity)

        if latest_saved_record:
            # the latest saved timestamp
            latest_saved_timestamp = latest_saved_record.timestamp
        else:
            # the list date
            latest_saved_timestamp = entity.timestamp

        if not latest_saved_timestamp:
            return None, None, self.default_size, None

        size = evaluate_size_from_timestamp(start_timestamp=latest_saved_timestamp, level=self.level,
                                            one_day_trading_minutes=self.one_day_trading_minutes)

        if self.start_timestamp:
            start = max(self.start_timestamp, latest_saved_timestamp)
        else:
            start = latest_saved_timestamp

        return start, None, size, None


class TimestampsDataRecorder(TimeSeriesDataRecorder):

    def __init__(self,
                 force_update=False,
                 sleeping_time=5,
                 exchanges=None,
                 entity_ids=None,
                 code=None,
                 codes=None,
                 day_data=False,
                 entity_filters=None,
                 ignore_failed=True,
                 real_time=False,
                 fix_duplicate_way='add',
                 start_timestamp=None,
                 end_timestamp=None) -> None:
        super().__init__(force_update, sleeping_time, exchanges, entity_ids, code=code, codes=codes, day_data=day_data,
                         entity_filters=entity_filters, ignore_failed=ignore_failed, real_time=real_time,
                         fix_duplicate_way=fix_duplicate_way, start_timestamp=start_timestamp,
                         end_timestamp=end_timestamp)
        self.security_timestamps_map = {}

    def init_timestamps(self, entity_item) -> List[pd.Timestamp]:
        raise NotImplementedError

    def evaluate_start_end_size_timestamps(self, entity):
        timestamps = self.security_timestamps_map.get(entity.id)
        if not timestamps:
            timestamps = self.init_timestamps(entity)
            if self.start_timestamp:
                timestamps = [t for t in timestamps if t >= self.start_timestamp]

            if self.end_timestamp:
                timestamps = [t for t in timestamps if t <= self.end_timestamp]

            self.security_timestamps_map[entity.id] = timestamps

        if not timestamps:
            return None, None, 0, timestamps

        timestamps.sort()

        self.logger.info(
            'entity_id:{},timestamps start:{},end:{}'.format(entity.id, timestamps[0], timestamps[-1]))

        latest_record = self.get_latest_saved_record(entity=entity)

        if latest_record:
            self.logger.info('latest record timestamp:{}'.format(latest_record.timestamp))
            timestamps = [t for t in timestamps if t >= latest_record.timestamp]

            if timestamps:
                return timestamps[0], timestamps[-1], len(timestamps), timestamps
            return None, None, 0, None

        return timestamps[0], timestamps[-1], len(timestamps), timestamps


# the __all__ is generated
__all__ = ['Meta', 'Recorder', 'EntityEventRecorder', 'TimeSeriesDataRecorder', 'FixedCycleDataRecorder',
           'TimestampsDataRecorder']


# -*- coding: utf-8 -*-
import enum


class BlockCategory(enum.Enum):
    # 行业版块
    industry = 'industry'
    # 概念版块
    concept = 'concept'
    # 区域版块
    area = 'area'


class IndexCategory(enum.Enum):
    # 中国指数提供商：
    # 中证指数公司 http://www.csindex.com.cn/zh-CN
    # 上证指数(上交所标的) 中证指数(沪深)

    # 国证指数公司 http://www.cnindex.com.cn/index.html
    # 深证指数(深交所标的) 国证指数(沪深)

    # 规模指数
    # 常见的上证指数，深证指数等
    scope = 'scope'
    # 行业指数
    industry = 'industry'
    # 风格指数
    style = 'style'
    # 主题指数
    #
    # 策略指数
    #
    # 综合指数
    #
    # 债券指数
    #
    # 基金指数
    fund = 'fund'
    # 定制指数
    #
    # 人民币指数
    #
    # 跨境指数
    #
    # 其他指数


class ReportPeriod(enum.Enum):
    # 有些基金的2，4季报只有10大持仓，半年报和年报有详细持仓，需要区别对待
    season1 = 'season1'
    season2 = 'season2'
    season3 = 'season3'
    season4 = 'season4'
    half_year = 'half_year'
    year = 'year'


# 用于区分不同的财务指标
class CompanyType(enum.Enum):
    qiye = 'qiye'
    baoxian = 'baoxian'
    yinhang = 'yinhang'
    quanshang = 'quanshang'


# the __all__ is generated
__all__ = ['BlockCategory', 'IndexCategory', 'ReportPeriod', 'CompanyType']

# __init__.py structure:
# common code of the package
# export interface in __all__ which contains __all__ of its sub modules

# import all from submodule misc
from .misc import *
from .misc import __all__ as _misc_all
__all__ += _misc_all

# import all from submodule quotes
from .quotes import *
from .quotes import __all__ as _quotes_all
__all__ += _quotes_all

# import all from submodule meta
from .meta import *
from .meta import __all__ as _meta_all
__all__ += _meta_all

# import all from submodule fundamental
from .fundamental import *
from .fundamental import __all__ as _fundamental_all
__all__ += _fundamental_all

# import all from submodule macro
from .macro import *
from .macro import __all__ as _macro_all
__all__ += _macro_all

# import all from submodule trader_info
from .trader_info import *
from .trader_info import __all__ as _trader_info_all
__all__ += _trader_info_all

# import all from submodule actor
from .actor import *
from .actor import __all__ as _actor_all
__all__ += _actor_all


# -*- coding: utf-8 -*-
import operator
from itertools import accumulate
from typing import List, Union, Type

import pandas as pd

from zvtm.contract import IntervalLevel, Mixin, TradableEntity
from zvtm.contract.factor import Factor, Transformer, Accumulator, FilterFactor
from zvtm.domain import FinanceFactor, BalanceSheet, Stock


class FinanceBaseFactor(Factor):
    def __init__(self, data_schema: Type[Mixin] = FinanceFactor, entity_schema: Type[TradableEntity] = Stock,
                 provider: str = None,
                 entity_provider: str = None, entity_ids: List[str] = None, exchanges: List[str] = None,
                 codes: List[str] = None, start_timestamp: Union[str, pd.Timestamp] = None,
                 end_timestamp: Union[str, pd.Timestamp] = None, columns: List = None, filters: List = None,
                 order: object = None, limit: int = None, level: Union[str, IntervalLevel] = None,
                 category_field: str = 'entity_id', time_field: str = 'timestamp', computing_window: int = None,
                 keep_all_timestamp: bool = False, fill_method: str = 'ffill', effective_number: int = None,
                 transformer: Transformer = None, accumulator: Accumulator = None, need_persist: bool = False,
                 only_compute_factor: bool = False, factor_name: str = None, clear_state: bool = False,
                 only_load_factor: bool = False) -> None:
        if not columns:
            columns = data_schema.important_cols()
        super().__init__(data_schema, entity_schema, provider, entity_provider, entity_ids, exchanges, codes,
                         start_timestamp, end_timestamp, columns, filters, order, limit, level, category_field,
                         time_field, computing_window, keep_all_timestamp, fill_method, effective_number, transformer,
                         accumulator, need_persist, only_compute_factor, factor_name, clear_state, only_load_factor)


class GoodCompanyFactor(FinanceBaseFactor, FilterFactor):
    def __init__(self,
                 data_schema: Type[Mixin] = FinanceFactor,
                 entity_schema: TradableEntity = Stock,
                 provider: str = None,
                 entity_provider: str = None,
                 entity_ids: List[str] = None,
                 exchanges: List[str] = None,
                 codes: List[str] = None,
                 start_timestamp: Union[str, pd.Timestamp] = None,
                 end_timestamp: Union[str, pd.Timestamp] = None,
                 columns: List = (FinanceFactor.roe,
                                  FinanceFactor.op_income_growth_yoy,
                                  FinanceFactor.net_profit_growth_yoy,
                                  FinanceFactor.report_period,
                                  FinanceFactor.op_net_cash_flow_per_op_income,
                                  FinanceFactor.sales_net_cash_flow_per_op_income,
                                  FinanceFactor.current_ratio,
                                  FinanceFactor.debt_asset_ratio),
                 filters: List = (FinanceFactor.roe >= 0.02,
                                  FinanceFactor.op_income_growth_yoy >= 0.05,
                                  FinanceFactor.net_profit_growth_yoy >= 0.05,
                                  FinanceFactor.op_net_cash_flow_per_op_income >= 0.1,
                                  FinanceFactor.sales_net_cash_flow_per_op_income >= 0.3,
                                  FinanceFactor.current_ratio >= 1,
                                  FinanceFactor.debt_asset_ratio <= 0.5
                                  ),
                 order: object = None,
                 limit: int = None,
                 level: Union[str, IntervalLevel] = IntervalLevel.LEVEL_1DAY,
                 category_field: str = 'entity_id',
                 time_field: str = 'timestamp',
                 computing_window: int = None,
                 keep_all_timestamp: bool = True,
                 fill_method: str = 'ffill',
                 effective_number: int = None,
                 transformer: Transformer = None,
                 accumulator: Accumulator = None,
                 need_persist: bool = False,
                 only_compute_factor: bool = False,
                 factor_name: str = None,
                 clear_state: bool = False,
                 only_load_factor: bool = False,
                 window='1095d',
                 count=8,
                 col_period_threshold={'roe': 0.02}) -> None:
        self.window = window
        self.count = count

        # 对于根据年度计算才有意义的指标，比如roe,我们会对不同季度的值区别处理,传入的参数为季度值
        self.col_period_threshold = col_period_threshold
        if self.col_period_threshold:
            if 'report_period' not in columns and (data_schema.report_period not in columns):
                columns.append(data_schema.report_period)

        self.logger.info(f'using data_schema:{data_schema.__name__}')

        super().__init__(data_schema, entity_schema, provider, entity_provider, entity_ids, exchanges, codes,
                         start_timestamp, end_timestamp, columns, filters, order, limit, level, category_field,
                         time_field, computing_window, keep_all_timestamp, fill_method, effective_number, transformer,
                         accumulator, need_persist, only_compute_factor, factor_name, clear_state, only_load_factor)

    def compute_factor(self):
        def filter_df(df):
            se = pd.Series(index=df.index)
            for index, row in df.iterrows():

                if row.report_period == 'year':
                    mul = 4
                elif row.report_period == 'season3':
                    mul = 3
                elif row.report_period == 'half_year':
                    mul = 2
                else:
                    mul = 1

                filters = []
                for col in self.col_period_threshold:
                    col_se = eval(f'row.{col}')
                    filters.append(col_se >= mul * self.col_period_threshold[col])
                se[index] = list(accumulate(filters, func=operator.__and__))[-1]

            return se

        if self.col_period_threshold:
            self.factor_df = self.data_df.loc[lambda df: filter_df(df), :]

        self.factor_df = pd.DataFrame(index=self.data_df.index, columns=['count'], data=1)

        self.factor_df = self.factor_df.reset_index(level=1)

        self.factor_df = self.factor_df.groupby(level=0).rolling(window=self.window, on=self.time_field).count()

        self.factor_df = self.factor_df.reset_index(level=0, drop=True)
        self.factor_df = self.factor_df.set_index(self.time_field, append=True)

        self.factor_df = self.factor_df.loc[(slice(None), slice(self.start_timestamp, self.end_timestamp)), :]

        self.logger.info('factor:{},factor_df:\n{}'.format(self.factor_name, self.factor_df))

    def compute_result(self):
        self.result_df = self.factor_df.apply(lambda x: x >= self.count)

        self.logger.info('factor:{},result_df:\n{}'.format(self.factor_name, self.result_df))


if __name__ == '__main__':
    # f1 = GoodCompanyFactor(keep_all_timestamp=False)
    # print(f1.result_df)

    # 高股息 低应收
    factor2 = GoodCompanyFactor(data_schema=BalanceSheet, columns=[BalanceSheet.accounts_receivable], filters=[
        BalanceSheet.accounts_receivable <= 0.2 * BalanceSheet.total_current_assets], keep_all_timestamp=False,
                                col_period_threshold=None)
    print(factor2.result_df)
# the __all__ is generated
__all__ = ['FinanceBaseFactor', 'GoodCompanyFactor']


# -*- coding: utf-8 -*-

import pandas as pd
from jqdatapy.api import get_token, get_bars

from zvtm import zvt_config
from zvtm.api.kdata import generate_kdata_id, get_kdata_schema, get_kdata
from zvtm.contract import IntervalLevel, AdjustType
from zvtm.contract.api import df_to_db
from zvtm.contract.recorder import FixedCycleDataRecorder
from zvtm.domain import Stock, StockKdataCommon, Stock1wkHfqKdata
from zvtm.recorders.joinquant.common import to_jq_trading_level, to_jq_entity_id
from zvtm.utils.pd_utils import pd_is_not_null
from zvtm.utils.time_utils import to_time_str, now_pd_timestamp, TIME_FORMAT_DAY, TIME_FORMAT_ISO8601


class JqChinaStockKdataRecorder(FixedCycleDataRecorder):
    entity_provider = 'joinquant'
    entity_schema = Stock

    # 数据来自jq
    provider = 'joinquant'

    # 只是为了把recorder注册到data_schema
    data_schema = StockKdataCommon

    def __init__(self, force_update=True, sleeping_time=10, exchanges=None, entity_ids=None,
                 code=None, codes=None, day_data=False,
                 entity_filters=None, ignore_failed=True, real_time=False, fix_duplicate_way='ignore',
                 start_timestamp=None, end_timestamp=None, level=IntervalLevel.LEVEL_1DAY, kdata_use_begin_time=False,
                 one_day_trading_minutes=24 * 60, adjust_type=AdjustType.qfq) -> None:
        level = IntervalLevel(level)
        adjust_type = AdjustType(adjust_type)
        self.data_schema = get_kdata_schema(entity_type='stock', level=level, adjust_type=adjust_type)
        self.jq_trading_level = to_jq_trading_level(level)

        super().__init__(force_update, sleeping_time, exchanges, entity_ids, code, codes, day_data, entity_filters,
                         ignore_failed, real_time, fix_duplicate_way, start_timestamp, end_timestamp, level,
                         kdata_use_begin_time, one_day_trading_minutes)

        self.adjust_type = adjust_type

        get_token(zvt_config['jq_username'], zvt_config['jq_password'], force=True)

    def init_entities(self):
        super().init_entities()
        # 过滤掉退市的
        self.entities = [entity for entity in self.entities if
                         (entity.end_date is None) or (entity.end_date > now_pd_timestamp())]

    def generate_domain_id(self, entity, original_data):
        return generate_kdata_id(entity_id=entity.id, timestamp=original_data['timestamp'], level=self.level)

    def recompute_qfq(self, entity, qfq_factor, last_timestamp):
        # 重新计算前复权数据
        if qfq_factor != 0:
            kdatas = get_kdata(provider=self.provider, entity_id=entity.id, level=self.level.value,
                               order=self.data_schema.timestamp.asc(),
                               return_type='domain',
                               session=self.session,
                               filters=[self.data_schema.timestamp < last_timestamp])
            if kdatas:
                self.logger.info('recomputing {} qfq kdata,factor is:{}'.format(entity.code, qfq_factor))
                for kdata in kdatas:
                    kdata.open = round(kdata.open * qfq_factor, 2)
                    kdata.close = round(kdata.close * qfq_factor, 2)
                    kdata.high = round(kdata.high * qfq_factor, 2)
                    kdata.low = round(kdata.low * qfq_factor, 2)
                self.session.add_all(kdatas)
                self.session.commit()

    def record(self, entity, start, end, size, timestamps):
        if self.adjust_type == AdjustType.hfq:
            fq_ref_date = '2000-01-01'
        else:
            fq_ref_date = to_time_str(now_pd_timestamp())

        if not self.end_timestamp:
            df = get_bars(to_jq_entity_id(entity),
                          count=size,
                          unit=self.jq_trading_level,
                          # fields=['date', 'open', 'close', 'low', 'high', 'volume', 'money'],
                          fq_ref_date=fq_ref_date)
        else:
            end_timestamp = to_time_str(self.end_timestamp)
            df = get_bars(to_jq_entity_id(entity),
                          count=size,
                          unit=self.jq_trading_level,
                          # fields=['date', 'open', 'close', 'low', 'high', 'volume', 'money'],
                          end_date=end_timestamp,
                          fq_ref_date=fq_ref_date)
        if pd_is_not_null(df):
            df['name'] = entity.name
            df.rename(columns={'money': 'turnover', 'date': 'timestamp'}, inplace=True)

            df['entity_id'] = entity.id
            df['timestamp'] = pd.to_datetime(df['timestamp'])
            df['provider'] = 'joinquant'
            df['level'] = self.level.value
            df['code'] = entity.code

            # 判断是否需要重新计算之前保存的前复权数据
            if self.adjust_type == AdjustType.qfq:
                check_df = df.head(1)
                check_date = check_df['timestamp'][0]
                current_df = get_kdata(entity_id=entity.id, provider=self.provider, start_timestamp=check_date,
                                       end_timestamp=check_date, limit=1, level=self.level,
                                       adjust_type=self.adjust_type)
                if pd_is_not_null(current_df):
                    old = current_df.iloc[0, :]['close']
                    new = check_df['close'][0]
                    # 相同时间的close不同，表明前复权需要重新计算
                    if round(old, 2) != round(new, 2):
                        qfq_factor = new / old
                        last_timestamp = pd.Timestamp(check_date)
                        self.recompute_qfq(entity, qfq_factor=qfq_factor, last_timestamp=last_timestamp)

            def generate_kdata_id(se):
                if self.level >= IntervalLevel.LEVEL_1DAY:
                    return "{}_{}".format(se['entity_id'], to_time_str(se['timestamp'], fmt=TIME_FORMAT_DAY))
                else:
                    return "{}_{}".format(se['entity_id'], to_time_str(se['timestamp'], fmt=TIME_FORMAT_ISO8601))

            df['id'] = df[['entity_id', 'timestamp']].apply(generate_kdata_id, axis=1)

            df = df.drop_duplicates(subset='id', keep='last')

            df_to_db(df=df, data_schema=self.data_schema, provider=self.provider, force_update=self.force_update)

        return None


if __name__ == '__main__':
    Stock1wkHfqKdata.record_data(codes=['300999'])

# the __all__ is generated
__all__ = ['JqChinaStockKdataRecorder']


# -*- coding: utf-8 -*-
import logging

import pandas as pd
import requests

from zvtm.domain import IndexCategory
from zvtm.recorders.consts import DEFAULT_HEADER
from zvtm.utils import to_pd_timestamp

logger = logging.getLogger(__name__)

original_page_url = 'http://www.csindex.com.cn/zh-CN/indices/index?class_2=2&class_17=17&class_10=10&class_7=7&is_custom_0=1'

url = 'http://www.csindex.com.cn/zh-CN/indices/index?page=1&page_size=1000&by=asc&order=%E5%8F%91%E5%B8%83%E6%97%B6%E9%97%B4&data_type=json' \
      '&{}&class_7=7&class_10=10&{}&is_custom_0=1'

# 中证指数 抓取 风格指数 行业指数 规模指数
csi_category_map_url = {
    IndexCategory.style: url.format('class_1=1', 'class_19=19'),
    IndexCategory.industry: url.format('class_1=1', 'class_18=18'),
    IndexCategory.scope: url.format('class_1=1', 'class_17=17'),
}

# 上证指数 只取规模指数
sh_category_map_url = {
    IndexCategory.scope: url.format('class_2=2', 'class_17=17')
}


def _get_resp_data(resp: requests.Response):
    resp.raise_for_status()
    return resp.json()['list']


def get_cs_index(index_type='sh', category=IndexCategory.scope):
    if index_type == 'csi':
        category_map_url = csi_category_map_url
    elif index_type == 'sh':
        category_map_url = sh_category_map_url
    else:
        logger.warning(f'not support index type: {index_type}')
        assert False

    url = category_map_url.get(category)

    requests_session = requests.Session()

    resp = requests_session.get(url, headers=DEFAULT_HEADER)

    results = _get_resp_data(resp)
    the_list = []

    logger.info(f'category: {category} ')
    logger.info(f'results: {results} ')
    for i, result in enumerate(results):
        logger.info(f'to {i}/{len(results)}')
        code = result['index_code']
        name = result['indx_sname']
        entity_id = f'index_sh_{code}'
        index_item = {
            'id': entity_id,
            'entity_id': entity_id,
            'timestamp': to_pd_timestamp(result['base_date']),
            'entity_type': 'index',
            'exchange': 'sh',
            'code': code,
            'name': name,
            'category': category.value,
            'list_date': to_pd_timestamp(result['online_date']),
            'base_point': result['base_point'],
            'publisher': 'csindex'
        }
        logger.info(index_item)
        the_list.append(index_item)
    if the_list:
        return pd.DataFrame.from_records(the_list)


if __name__ == '__main__':
    df = get_cs_index()
    print(df)
# the __all__ is generated
__all__ = ['get_cs_index']


# -*- coding: utf-8 -*-
import subprocess


def get_git_user_name():
    try:
        return subprocess.check_output(['git', 'config', '--get', 'user.name']).decode('utf8').strip()
    except:
        return "foolcage"


def get_git_user_email():
    try:
        return subprocess.check_output(['git', 'config', '--get', 'user.email']).decode('utf8').strip()
    except:
        return ""
# the __all__ is generated
__all__ = ['get_git_user_name', 'get_git_user_email']


# -*- coding: utf-8 -*-

from sqlalchemy.orm import declarative_base

from zvtm.contract import TradableEntity
from zvtm.contract.register import register_schema, register_entity

StockusMetaBase = declarative_base()


# 美股
@register_entity(entity_type='stockus')
class Stockus(StockusMetaBase, TradableEntity):
    __tablename__ = 'stockus'


register_schema(providers=['em'], db_name='stockus_meta',
                schema_base=StockusMetaBase)
# the __all__ is generated
__all__ = ['Stockus']


# -*- coding: utf-8 -*-
from sqlalchemy.ext.declarative import declarative_base

from zvtm.contract.register import register_schema
from zvtm.factors.ma.domain.common import MaStatsFactorCommon

Stock1dMaStatsFactorBase = declarative_base()


class Stock1dMaStatsFactor(Stock1dMaStatsFactorBase, MaStatsFactorCommon):
    __tablename__ = 'stock_1d_ma_stats_factor'


register_schema(providers=['zvt'], db_name='stock_1d_ma_stats_factor', schema_base=Stock1dMaStatsFactorBase)
# the __all__ is generated
__all__ = ['Stock1dMaStatsFactor']


# -*- coding: utf-8 -*-
from zvtm.contract import IntervalLevel
from zvtm.factors.algorithm import MaTransformer, MacdTransformer
from zvtm.factors.ma.ma_factor import CrossMaFactor
from zvtm.domain import *
def getdata():
    Stock.record_data(provider='joinquant')
    # Stock.record_data(provider='eastmoney')
    # FinanceFactor.record_data(provider='joinquant')
    # BalanceSheet.record_data(provider='joinquant')
    # IncomeStatement.record_data(provider='joinquant')
    # CashFlowStatement.record_data(provider='joinquant')


# -*- coding: utf-8 -*-
# the __all__ is generated
__all__ = []

# __init__.py structure:
# common code of the package
# export interface in __all__ which contains __all__ of its sub modules

# import all from submodule overall
from .overall import *
from .overall import __all__ as _overall_all
__all__ += _overall_all

# import all from submodule money_flow
from .money_flow import *
from .money_flow import __all__ as _money_flow_all
__all__ += _money_flow_all

# import all from submodule holder
from .holder import *
from .holder import __all__ as _holder_all
__all__ += _holder_all